/*
 * This is part of the doxml library, a parser for XML.  It should not
 *  require any functionality not defined in ANSI C's standard libraries.
 *
 * Copyright (C) 1999 John Stracke, http://www.thibault.org
 *
 * This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Log: parse.c,v $
 * Revision 1.50  1999/08/29 17:42:59  francis
 * Fixing a debug message in get_attlist() to (a) name the right
 * function, and (b) not complain if the doc has no DOCTYPE.
 *
 * Revision 1.49  1999/07/31 15:16:47  francis
 * Adding an error message
 *
 * Revision 1.48  1999/07/24 22:41:03  francis
 * Pseudo-fixing a memory corruption problem, at the cost of memory leaks
 *
 * Revision 1.47  1999/07/24 22:29:15  francis
 * Fixing some memory errors
 *
 * Revision 1.46  1999/07/24 21:58:30  francis
 * Implementing late binding for entity values, so that an entity can
 * reference another entity which is defined later.
 *
 * Revision 1.45  1999/07/24 20:23:51  francis
 * Adding support for param entities in the DTD
 *
 * Revision 1.44  1999/07/24 19:13:31  francis
 * Adding support for notations
 *
 * Revision 1.43  1999/07/24 16:55:28  francis
 * Adding support for entity-valued attributes
 *
 * Revision 1.42  1999/07/24 15:54:04  francis
 * Adding support for parameter entites in entity definitions.
 *
 * Revision 1.41  1999/07/24 15:37:35  francis
 * Converting get_attlist() and get_attr_def() to use the new
 * doxml_tags_match_dtd()
 *
 * Revision 1.40  1999/07/24 15:23:03  francis
 * Adding support for default attribute values
 *
 * Revision 1.39  1999/07/24 14:53:14  francis
 * Adding support for attribute value normalization
 *
 * Revision 1.38  1999/07/24 13:43:46  francis
 * Adding support for namespaces in entity names.
 *
 * Revision 1.37  1999/07/24 13:13:39  francis
 * Fixing the problem where <D:foo> can be terminated by </foo> instead of
 * by </D:foo>.
 *
 * Revision 1.36  1999/07/19 01:31:12  francis
 * Implementing namespace reform (but not using it yet).  Not documented yet.
 *
 * Revision 1.35  1999/07/18 22:36:58  francis
 * Adding support for character entities
 *
 * Revision 1.34  1999/07/18 20:25:06  francis
 * Moving the global buff (static in parse.c) into the doxml_context,
 * to make the parser reentrant, so that it can be used from two threads
 * simultaneously.  Documenting the change; updating the version number
 * of the manual.
 *
 * Revision 1.33  1999/07/08 14:34:52  francis
 * Documenting compliance with the entity processing table in section 4.4
 * of the spec.  To this end, adding more test cases in tst.xml, and
 * making a small change to parse.c to forbid unparsed entities in
 * content.
 *
 * Revision 1.32  1999/07/08 13:52:22  francis
 * Adding support for external parsed entities (I don't parse them yet,
 * but I do include an entity atom).  I'll hook in the URI framework
 * later.
 *
 * Fixing a bug in parse_markupdecl() where entities with no NDATA didn't
 * zero out notation_name.
 *
 * Revision 1.31  1999/06/25 21:15:00  francis
 * Adding a framework for the application to provide doxml with URI
 * handlers for use in processing external IDs.  (Not providing URI
 * code as part of doxml means we can work with whatever URI code the
 * app wants, and don't have to write anything non-portable.)
 *
 * Also, changing old _LIBXML_ stuff in the headers to _DOXML_.
 *
 * Revision 1.30  1999/06/25 19:43:03  francis
 * Adding support for unparsed entities
 *
 * Revision 1.29  1999/06/25 18:48:40  francis
 * Adding support for internal parsed entities
 *
 * Revision 1.28  1999/06/21 03:23:51  francis
 * Adding support for comments
 *
 * Revision 1.27  1999/06/21 03:19:56  francis
 * Finishing up <!ENTITY> support
 *
 * Revision 1.26  1999/06/21 02:53:28  francis
 * Adding the ability to parse <!ENTITY>s.  Can't write them yet, though.
 *
 * Revision 1.25  1999/06/21 02:47:02  francis
 * Converted parse_external_id() to use parse_quoted()
 *
 * Revision 1.24  1999/06/21 02:42:12  francis
 * Converted parse_element() to use parse_quoted()
 *
 * Revision 1.23  1999/06/21 02:40:15  francis
 * Doing some work on parsing entity decls.
 * Checking in out of paranoia.
 * Added a parse_quote(); trying to use it elsewhere.
 *
 * Revision 1.22  1999/06/21 02:11:00  francis
 * Making sure default values in an <!ATTLIST> expand their & entities.
 * Also, in making those entities print out right, generalized the
 * escaping used in printing normal attributes
 *
 * Revision 1.21  1999/06/21 02:02:32  francis
 * Converting parse_entity() to return a string.  This is in preparation for
 * DTD-defined entities.
 *
 * Revision 1.20  1999/06/21 00:59:31  francis
 * Adding support for <!ATTLIST>s
 *
 * Revision 1.19  1999/06/20 22:23:54  francis
 * Adding support for <!ELEMENT>s in the doctype
 *
 * Revision 1.18  1999/06/20 20:05:34  francis
 * Adding support for PIs in the doctype
 *
 * Revision 1.17  1999/06/20 16:30:14  francis
 * Adding partial support for <!DOCTYPE>.  It now handles doctypes with
 * names (all doctypes have names), with external-ids, but not with
 * any of the declarations.
 *
 * Revision 1.16  1999/06/13 04:05:57  francis
 * Improving error reporting, so that the parser actually returns NULL when
 * given a non-well-formed document.
 *
 * Revision 1.15  1999/06/13 03:44:28  francis
 * Adding support for comments
 *
 * Revision 1.14  1999/06/13 02:15:40  francis
 * Adding support for CDATA
 *
 * Revision 1.13  1999/06/08 02:45:57  francis
 * Adding support for & entities in attribute values
 *
 * Revision 1.12  1999/06/08 02:39:04  francis
 * Updating comment listing which entities are supported
 *
 * Revision 1.11  1999/06/08 02:38:08  francis
 * Adding support for &apos; and &quot;
 *
 * Revision 1.10  1999/06/07 04:57:38  francis
 * Implementing processing instructions.  Note the API changes.
 *
 * Revision 1.9  1999/06/01 03:42:06  francis
 * Adding support for the xmlns="uri" form of namespace specifiers.
 *
 * Revision 1.8  1999/06/01 02:44:41  francis
 * Changing doxml_tags_match() so that namespaces come before tags, as in
 * other functions.
 *
 * Revision 1.7  1999/06/01 02:23:47  francis
 * Adding support for single quotes on attribute values
 *
 * Revision 1.6  1999/06/01 01:51:14  francis
 * Adding functions to get/set/clear attributes.  Reorganzing the
 * to-do list.  Fixing a parser bug introduced I removed support for
 * attributes with no values at release 0.2.
 *
 * Revision 1.5  1999/05/31 14:43:31  francis
 * Removing the support for attributes without values (which turn out to
 * be illegal).
 *
 * Revision 1.4  1999/05/21 23:10:28  francis
 * Reformatting, since renaming to doxml changed the widths of the
 * identifiers.
 *
 * Revision 1.3  1999/05/21 23:05:42  francis
 * Finished renaming to doxml
 *
 * Revision 1.2  1999/05/21 22:59:48  francis
 * Partially renaming to doxml
 *
 * Revision 1.1.1.1  1999/05/21 22:20:36  francis
 * Initial checkin of libxml, to be renamed doxml
 *
 */

static const char* id="$Id: parse.c,v 1.50 1999/08/29 17:42:59 francis Exp $";

/*#undef DEBUG*/

#include <string.h>

#include <doxml/priv.h>
#include <doxml/utf8.h>
#include <doxml/buff.h>

#include <ctype.h>
#include <stdlib.h>

static doxml_atom*
doxml_parse_atom(doxml_context* context,
                 doxml_atom* parent,
                 doxml_doctype** doctype_out,
                 const doxml_doctype* doctype_in,
                 int* null_means_error);
static doxml_atom*
parse_pi(doxml_context* context,
         doxml_atom* parent);
static int
expect_poly(doxml_context* context,
            const char* choices[]);
static int
expect(doxml_context* context,
       const char* s);

static doxml_markupdecl*
get_entity(const doxml_name_in_dtd* entity_name,
           const doxml_doctype* doctype);

static const char*
safestr(const char* s)
{
  return (s ? s : "(null)");
}

static doxml_atom*
new_atom(doxml_atom* parent,
         doxml_atom_type type)
{
  doxml_atom* res;

#ifdef DEBUG
  fprintf(stderr,"doxml: new_atom()\n");
#endif

  res=(doxml_atom*)malloc(sizeof(doxml_atom));
  if (!res)
    {
      fprintf(stderr,"doxml: new_atom(): malloc() failed\n");
      return 0;
    }

  res->type=type;
  memset(&(res->data),0,sizeof(res->data));

  res->parent=parent;
  if (parent)
    {
      if (parent->data.element.atoms.last)
        {
          parent->data.element.atoms.last->next=res;
          res->prev=parent->data.element.atoms.last;
          parent->data.element.atoms.last=res;
          res->next=0;
        }
      else
        {
          parent->data.element.atoms.last=parent->data.element.atoms.first=res;
          res->next=res->prev=0;
        }
    }
  else
    res->next=res->prev=0;

  return res;
}

static doxml_atom*
new_text_atom(doxml_atom* parent,
              const char* text)
{
  doxml_atom* res;

#ifdef DEBUG
  fprintf(stderr,"doxml: new_text_atom(%s)\n",
          safestr(text));
#endif

  if (!text)
    {
      fprintf(stderr,"doxml: new_text_atom(): text is NULL\n");
      return 0;
    }

  res=new_atom(parent,doxml_atom_text);
  if (!res)
    {
      fprintf(stderr,"doxml: new_text_atom(): new_atom() failed\n");
      return 0;
    }

  res->data.text.text=text;
  return res;
}

static doxml_atom*
new_element_atom(doxml_atom* parent,
                 const char* namespace,
                 const char* prefix,
                 const char* tag)
{
  doxml_atom* res;

#ifdef DEBUG
  fprintf(stderr,"doxml: new_element_atom(%s,%s)\n",
          safestr(namespace),safestr(tag));
#endif

  if (!tag)
    {
      fprintf(stderr,"doxml: new_element_atom(): tag is NULL\n");
      return 0;
    }

  if (namespace)
    {
      namespace=strdup(namespace);
      if (!namespace)
        {
          fprintf(stderr,"doxml: new_element_atom(): could not strdup(namespace)\n");
          return 0;
        }
    }

  res=new_atom(parent,doxml_atom_element);
  if (!res)
    {
      fprintf(stderr,"doxml: new_element_atom(): new_atom() failed\n");
      if (namespace)
        free((char*)namespace); /* cast away const */
      return 0;
    }

  res->data.element.name.space=namespace;
  res->data.element.name.prefix=prefix;
  res->data.element.name.name=tag;
  return res;
}

static doxml_atom*
new_pi_atom(doxml_atom* parent,
            const char* space,
            const char* prefix,
            const char* name,
            const char* value)
{
  doxml_atom* res;

#ifdef DEBUG
  fprintf(stderr,"doxml: new_pi_atom(%s,%s)\n",
          safestr(name),safestr(value));
#endif

  if (!name)
    {
      fprintf(stderr,"doxml: new_pi_atom(): name is NULL\n");
      return 0;
    }

  name=strdup(name);
  if (!name)
    {
      fprintf(stderr,"doxml: new_pi_atom(): cannot strdup(name)\n");
      return 0;
    }

  if (value)
    {
      value=strdup(value);
      if (!value)
        {        
          fprintf(stderr,"doxml: new_element_atom(): could not strdup(value)\n");
          return 0;
        }
    }

  res=new_atom(parent,doxml_atom_pi);
  if (!res)
    {
      fprintf(stderr,"doxml: new_element_atom(): new_atom() failed\n");
      if (name)
        free((char*)name); /* cast away const */
      if (value)
        free((char*)value); /* cast away const */
      return 0;
    }

  res->data.pi.name.name=name;
  res->data.pi.name.prefix=prefix;
  res->data.pi.name.space=space;
  res->data.pi.value=value;
  return res;
}

static doxml_atom*
new_external_entity_atom(doxml_atom* parent,
                         const doxml_markupdecl* decl)
{
  doxml_atom* res;

  if (!decl)
    {
      fprintf(stderr,"doxml: new_external_entity_atom(): NULL decl\n");
      return 0;
    }

  res=new_atom(parent,doxml_atom_external_entity);
  if (!res)
    {
      fprintf(stderr,"doxml: new_external_entity_atom(): new_atom() failed\n");
      return 0;
    }

  res->data.external_entity.decl=decl;
  return res;
}

static int
skip_ws(doxml_context* context,
        int must)
{
  int ch,N=0;
  while (isspace(ch=doxml_getc(context)))
    N++;

  if (must && (N==0))
    {
      fprintf(stderr,"doxml: %d: skip_ws(): didn't find required whitespace\n",
              context->line_number);
      return -2;
    }

  return ch;
}

/* can't use isalpha(), because it might be localized */
static int
is_letter(int ch)
{
  if (ch>127)
    return 1;
  if ((ch>='A')&&(ch<='Z'))
    return 1;
  if ((ch>='a')&&(ch<='z'))
    return 1;
  return 0;
}

static int
is_name_char(int ch)
{
  if (is_letter(ch))
    return 1;

  if ((ch>='0')&&(ch<='9'))
    return 1;

  switch (ch)
    {
    case '.':
    case '-':
    case '_':
    case ':':
      return 1;

    default:
      return 0;
    }
}

static const char*
parse_name_or_nmtoken(doxml_context* context,
                      int is_name)
{
  int ch;
  if (doxml_buff_init(context)<0)
    {
      fprintf(stderr,"doxml: parse_name_or_nmtoken(): buff_init() failed\n");
      return 0;
    }

  ch=doxml_getc(context);
  if (ch==EOF)
    {
      fprintf(stderr,"doxml: parse_name_or_nmtoken(): EOF at start of name\n");
      return 0;
    }

  if (is_name)
    {
      if (!(is_letter(ch) || (ch=='_') || (ch==':')))
        {
          fprintf(stderr,"doxml: parse_name_or_nmtoken(): name starts with illegal character '%c'\n",ch);
          return 0;
        }
    }
  else
    {
      if (!is_name_char(ch))
        {
          fprintf(stderr,"doxml: parse_name_or_nmtoken(): nmtoken starts with illegal character '%c'\n",ch);
          return 0;
        }
    }

  do {
    if (doxml_buff_append(context,ch)<0)
      {
        fprintf(stderr,"doxml: parse_name_or_nmtoken(): cannot append character\n");
        return 0;
      }

    ch=doxml_getc(context);
  } while (is_name_char(ch));

  if (ch==EOF)
    {
      fprintf(stderr,"doxml: parse_name_or_nmtoken(): EOF in name\n");
      return 0;
    }

  doxml_ungetc(ch,context);  

  {
    const char* name=doxml_buff_extract(context);
    if (!name)
      {
        fprintf(stderr,"doxml: parse_name_or_nmtoken(): can't extract name from buffer\n");
        return 0;
      }
    return name;
  }
}

static int
parse_name_in_dtd(doxml_context* context,
                  doxml_name_in_dtd* name)
{
  const char* colon;
  /* cast away const so we can free() it below */
  char* prefixedname;

  name->name=name->prefix=0;

  prefixedname=(char*)parse_name_or_nmtoken(context,1);
  if (!prefixedname)
    return -1;

#ifdef DEBUG
  fprintf(stderr,"doxml: parse_name_in_dtd(): prefixedname==%s\n",
          prefixedname);
#endif
  
  colon=strchr(prefixedname,':');
  if (colon)
    {
#ifdef DEBUG
      fprintf(stderr,"doxml: parse_name_in_dtd(): colon==%s\n",
              colon);
#endif
      
      if (!(colon[1]))
        {
          fprintf(stderr,"doxml: parse_name_in_dtd(): ':' at end of name \"%s\"\n",
                  prefixedname);
          free(prefixedname);
          return -1;
        }
      if (!(name->name=strdup(colon+1)))
        {
          fprintf(stderr,"doxml: parse_name_in_dtd(): can't strdup() for name\n");
          free(prefixedname);
          return -1;
        }
      {
        int len=colon-prefixedname;
        char* prefix=(char*)malloc(len+1);
        if (!prefix)
          {
            fprintf(stderr,"doxml: parse_name_in_dtd(): can't malloc() for prefix\n");
            free(prefixedname);
            return -1;
          }
        strncpy(prefix,prefixedname,len);
        prefix[len]=0;
        name->prefix=prefix;
        free(prefixedname);
      }
    }
  else
    {
      name->name=prefixedname;
      name->prefix=0;
    }
  return 0;
}

static int
parse_name(doxml_context* context,
           doxml_name* name)
{
  doxml_name_in_dtd dtd;
  if (parse_name_in_dtd(context,&dtd)<0)
    {
      fprintf(stderr,"doxml: parse_name(): parse_name_in_dtd() failed\n");
      return -1;
    }
  name->name=dtd.name;
  name->prefix=dtd.prefix;
  name->space=0;
  return 0;
}

static const char*
parse_nmtoken(doxml_context* context)
{
  return parse_name_or_nmtoken(context,0);
}

static int
hex_digit_to_int(char ch)
{
  if (ch<'0')
    return -1;
  if (ch<='9')
    return (ch-'0');
  ch=tolower(ch);
  if (ch<'a')
    return -1;
  if (ch<='f')
    return (ch-'a')+10;
  return -1;
}

static int
digit_to_int(char ch)
{
  if (ch<'0')
    return -1;
  if (ch<='9')
    return (ch-'0');
  return -1;
}

static const char*
eval_character_entity(const char* str)
{
  int value=0;
  if (!str)
    {
      fprintf(stderr,"doxml: eval_character_entity(): str is NULL\n");
      return 0;
    }

  if (str[0]=='x')
    {
      char ch;
      str++;
      while (ch=(*(str++)))
        {
          int digit;

          digit=hex_digit_to_int(ch);
          if (digit<0)
            {
              fprintf(stderr,
                      "doxml: eval_character_entity(): illegal hex digit\n");
              return 0;
            }
          value=16*value+digit;
        }
    }
  else
    {
      char ch;
      while (ch=(*(str++)))
        {
          int digit;

          digit=digit_to_int(ch);
          if (digit<0)
            {
              fprintf(stderr,
                      "doxml: eval_character_entity(): illegal digit\n");
              return 0;
            }
          value=10*value+digit;
        }
    }

  {
    char* res=malloc(10);
    doxml_unicode_char u;
    u.word=doxml_word_host_to_network_order(value);
    if (doxml_unicode_char_to_utf8(&u,res,10)<0)
      {
        fprintf(stderr,
                "doxml: eval_character_entity(): can't convert to UTF-8\n");
        free(res);
        return 0;
      }

    return res;
  }
}

static const char*
parse_character_entity(doxml_context* context)
{
  int ch;
  char str[40]; /* excessive */
  int len=0;

  while ((ch=doxml_getc(context))!=';')
    {
      if (ch==EOF)
        {
          fprintf(stderr,
                  "doxml: parse_character_entity(): EOF in reference\n");
          return 0;
        }

      if (len>(sizeof(str)-2))
        {
          fprintf(stderr,
                  "doxml: parse_character_entity(): reference too long\n"
                  );
          return 0;
        }

      str[len++]=ch;
    }

  str[len]=0;
  return eval_character_entity(str);
}

static const char*
parse_entity(doxml_context* context,
             const doxml_doctype* doctype,
             const doxml_markupdecl** decl,
             const char** notation_prefix,
             int* expandable,
             int* must_free_after)
{
  static const char* predefined_values[]={"&","'","\"","<",">"};
  int num_predefined_choices
    =sizeof(predefined_values)/sizeof(predefined_values[0]);
      
  int num_from_doctype=0;
  int num_choices=num_predefined_choices;
  const char** choices
    =(const char**)malloc(sizeof(const char*)*(num_choices+1));
  const char** choices_to_delete=0;
  int num_choices_to_delete=0;
  
  const doxml_markupdecl** entities=0;

  if (!choices)
    {
      fprintf(stderr,"doxml: parse_entity(): can't malloc() choices\n");
      return 0;
    }

  if (expandable)
    (*expandable)=0;

  choices_to_delete=(const char**)malloc(sizeof(const char*));
  if (!choices_to_delete)
    {
      fprintf(stderr,"doxml: parse_entity(): can't malloc() choices_to_delete\n");
      return 0;
    }
  choices_to_delete[0]=0;

  if (!must_free_after)
    {
      fprintf(stderr,"doxml: parse_entity(): must_free_after is NULL\n");
      return 0;
    }

  (*must_free_after)=0;
  if (decl)
    (*decl)=0;

  {
    int ch=doxml_getc(context);
    if (ch=='#')
      {
        const char* res=parse_character_entity(context);
        if (res)
          (*must_free_after)=1;
        return res;
      }

    if (doxml_ungetc(ch,context)<0)
      {
        fprintf(stderr,"doxml: parse_entity(): ungetc failed\n");
        return 0;
      }
  }

  choices[0]="amp";
  choices[1]="apos";
  choices[2]="quot";
  choices[3]="lt";
  choices[4]="gt";
  choices[num_choices]=0;

  if (doctype)
    {
      doxml_markupdecl* cur;
      for (cur=doctype->decls.first; cur; cur=cur->next)
        {
          if (cur->type!=doxml_markupdecl_entity)
            continue;

          if (!(cur->data.entity.name.name))
            continue;

          {
            const char** new_choices
              =(const char**)realloc(choices,
                                     sizeof(const char*)*(num_choices+2));
            if (!new_choices)
              {
                fprintf(stderr,"doxml: parse_entity(): can't realloc() choices for %s\n",
                        cur->data.entity.name.name);
                break;
              }

            choices=new_choices;
          }

          {
            const doxml_markupdecl** new_entities
              =(const doxml_markupdecl**)realloc(entities,
                                                 sizeof(doxml_markupdecl*)
                                                 *(num_from_doctype+1)
                                                 );
            if (!new_entities)
              {
                fprintf(stderr,"doxml: parse_entity(): can't realloc() entities for %s\n",
                        cur->data.entity.name.name);
                break;
              }

            entities=new_entities;
          }

          {
            char* name=0;
            if (cur->data.entity.name.prefix)
              {
                int len=strlen(cur->data.entity.name.name);
                len+=strlen(cur->data.entity.name.prefix);
                len++;

#ifdef DEBUG
                fprintf(stderr,"doxml: parse_entity(): adding %s:%s\n",
                        cur->data.entity.name.prefix,
                        cur->data.entity.name.name);
#endif

                {
                  const char** new_choices_to_delete
                    =(const char**)realloc(choices_to_delete,
                                           sizeof(const char*)*(num_choices_to_delete+2));
                  if (!new_choices_to_delete)
                    {
                      fprintf(stderr,"doxml: parse_entity(): can't realloc() choices_to_delete for %s\n",
                              cur->data.entity.name.name);
                      break;
                    }

                  choices_to_delete=new_choices_to_delete;
                }

                name=(char*)malloc(len+1);

                if (name)
                  {
                    sprintf(name,"%s:%s",
                            cur->data.entity.name.prefix,
                            cur->data.entity.name.name);
                    choices_to_delete[num_choices_to_delete++]=name;
                    choices_to_delete[num_choices_to_delete]=0;
                  }
              }
            else
              name=strdup(cur->data.entity.name.name);
            
            if (!name)
              {
                fprintf(stderr,"doxml: parse_entity(): can't allocate name\n");
                return 0;
              }

            choices[num_choices++]=name;
            choices[num_choices]=0;
          }            

          entities[num_from_doctype++]=cur;
        }
    }

  {
    int which=expect_poly(context,choices);
    if (   (which<0)
        || (which>num_choices) /* can't happen */
        )
      {
        fprintf(stderr,"doxml: parse_entity(): expect_poly() did not find any legal choice\n");
        if (entities)
          free(entities);
        {
          int i;
          for (i=0; i<num_choices_to_delete; i++)
            free((char*)(choices_to_delete[i]));
        }
        free(choices_to_delete);
        free(choices);
        choices=0;
        return 0;
      }

    if (expect(context,";")<0)
      {
        fprintf(stderr,"doxml: parse_entity(): entity %s not followed by ;\n",
                choices[which]);
        {
          int i;
          for (i=0; i<num_choices_to_delete; i++)
            free((char*)(choices_to_delete[i]));
        }
        free(choices_to_delete);
        free(choices);
        choices=0;
        return 0;
      }

    {
      int i;
      for (i=0; i<num_choices_to_delete; i++)
        free((char*)(choices_to_delete[i]));
    }
    free(choices_to_delete);
    free(choices);
    choices=0;

    if (which<num_predefined_choices)
      {
        if (entities)
          free(entities);
        
        return predefined_values[which];
      }

    if (!entities)
      {
        fprintf(stderr,"doxml: parse_entity(): expect_poly() found doctype-defined entity, but none are defined\n");
        return 0;
      }

    {
      const doxml_markupdecl* entity=entities[which-num_predefined_choices];
      const char* internal=0;
      free(entities);
      entities=0;

      if (!entity)
        {
          fprintf(stderr,
                  "doxml: parse_entity(): found NULL entity.  Can't happen\n");
          return 0;
        }

      if (expandable)
        (*expandable)=1;

      if (entity->data.entity.is_general)
        {
          if (entity->data.entity.data.general.is_external)
            {
              if (decl)
                {
                  (*decl)=entity;
                  if (notation_prefix)
                    (*notation_prefix)=entity->data.entity.data.general.data.external.notation_name.prefix;
                  return entity->data.entity.data.general.data.external.notation_name.name;
                }
              else
                {
                  fprintf(stderr,"doxml: parse_entity(): found external entity, but caller didn't want external\n");
                  return 0;
                }
            }
          return entity->data.entity.data.general.data.internal;
        }
      else
        {
          if (entity->data.entity.data.param.is_external)
            {
              if (decl)
                {
                  (*decl)=entity;
                  if (notation_prefix)
                    (*notation_prefix)=0;
                  return 0;
                }
              else
                {
                  fprintf(stderr,"doxml: parse_entity(): found external entity, but caller didn't want external\n");
                  return 0;
                }
            }
          return entity->data.entity.data.param.data.internal;
        }
    }
  }
}

static const char*
get_entity_value(const char* entity_name,
                 const doxml_doctype* doctype,
                 doxml_markupdecl** entity_out,
                 int* must_free_after)
{
  doxml_markupdecl* entity;
  doxml_name_in_dtd name;

  char* p=0;

  (*must_free_after)=0;
  if (entity_out)
    (*entity_out)=0;

  {
    const char* colon=strchr(entity_name,':');
    if (colon)
      {
        p=(char*)malloc(colon-entity_name+1);
        strncpy(p,entity_name,colon-entity_name);
        name.prefix=p;
        name.name=colon+1;
      }
    else
      {
        name.prefix=0;
        name.name=entity_name;
      }
  }

  if (!(name.prefix))
    {
      if (!strcmp(name.name,"amp"))
        return "&";
      if (!strcmp(name.name,"apos"))
        return "'";
      if (!strcmp(name.name,"quot"))
        return "\"";
      if (!strcmp(name.name,"lt"))
        return "<";
      if (!strcmp(name.name,"gt"))
        return ">";

      if (name.name[0]=='#')
        {
          const char* res=eval_character_entity(name.name+1);
          (*must_free_after)=1;
          return res;
        }
    }

  entity=get_entity(&name,doctype);
  if (entity_out)
    (*entity_out)=entity;

  if (p)
    free(p);

  if (!entity)
    {
      fprintf(stderr,"get_entity_value(): no such entity as %s\n",
              entity_name);
      return 0;
    }

  if (entity->data.entity.is_general)
    {
      if (entity->data.entity.data.general.parsed)
        return entity->data.entity.data.general.parsed;
      
      if (entity->data.entity.data.general.is_external)
        {
          fprintf(stderr,
                  "get_entity_value(): external entities not supported\n");
          return 0;
        }
      
      return entity->data.entity.data.general.data.internal;
    }
  else
    {
      if (entity->data.entity.data.param.is_external)
        {
          fprintf(stderr,
                  "get_entity_value(): external entities not supported\n");
          return 0;
        }
      
      return entity->data.entity.data.param.data.internal;
    }
}

static const char*
expand_entity_str(const char* entity_str,
                  const doxml_doctype* doctype,
                  int* must_free_expanded);

static const char*
get_entity_value_expanded(const char* entity_name,
                          const doxml_doctype* doctype,
                          int* must_free_after
                          )
{
  doxml_markupdecl* entity;
  const char* value=get_entity_value(entity_name,
                                     doctype,
                                     &entity,
                                     must_free_after);
  if (!value)
    {
      fprintf(stderr,"doxml: get_entity_value_expanded(): can't get entity value\n");
      return 0;
    }

  if (!strchr(value,'&'))
    return value;

  if (entity)
    {
      if (entity->data.entity.data.general.parsing)
        {
          fprintf(stderr,"doxml: get_entity_value_expanded(): circular reference detected\n");
          if (*must_free_after)
            free((char*)value);
          return 0;
        }

      entity->data.entity.data.general.parsing=1;
    }

  {
    int must_free_expanded;
    const char* expansion=expand_entity_str(value,doctype,
                                            &must_free_expanded);
    if (entity)
      entity->data.entity.data.general.parsing=0;
    if (!expansion)
      {
        fprintf(stderr,
                "doxml: get_entity_value_expanded(): can't expand\n");
        return value;
      }
    if (*must_free_after)
      free((char*)value);
    (*must_free_after)=must_free_expanded;
    return expansion;
  }
}

static char*
append_str(char* str,
           int* len,
           const char* extra,
           int extra_len)
{
  if (!(str && len && extra))
    return 0;

  if (extra_len==-1)
    extra_len=strlen(extra);

  if (!extra_len)
    return str;

  {
    char* new_str=(char*)realloc(str,extra_len+(*len)+1);
    if (!new_str)
      return 0;

    strncpy(new_str+(*len),extra,extra_len);
    (*len)+=extra_len;
    new_str[(*len)]=0;
    return new_str;
  }
}

static const char*
expand_entity_str(const char* entity_str,
                  const doxml_doctype* doctype,
                  int* must_free_expanded)
{
  char* res=(char*)malloc(1);
  int reslen=0;
  const char* amp;
  
  if (res)
    res[0]=0;
  else
    {
      fprintf(stderr,"doxml: expand_entity_str(): can't malloc initial res\n");
      return 0;
    }

  if (entity_str[0]=='&')
    {
      const char* semi=strchr(entity_str,';');
      /* Is the ; at the end of the string? */
      if (semi && (!semi[1]))
        {
          char* name=strdup(entity_str+1);
          name[semi-entity_str-1]=0;
          {
            const char* value=get_entity_value_expanded(name,
                                                        doctype,
                                                        must_free_expanded);
            free(name);
            if (!value)
              {
                fprintf(stderr,"doxml: expand_entity_str(): can't get value\n");
                (*must_free_expanded)=0;
                return 0;
              }

            return value;
          }
        }
    }

  (*must_free_expanded)=0;
  
  while (amp=strchr(entity_str,'&'))
    {
      {
        int len_before=amp-entity_str;
        if (len_before>0)
          {
            char* new_res=append_str(res,&reslen,entity_str,len_before);
            if (!new_res)
              {
                fprintf(stderr,"doxml: expand_entity_str(%s): can't append before amp\n",entity_str);
                if (res)
                  free(res);
                return 0;
              }
            res=new_res;
          }
      }
      
      {
        const char* semi=strchr(amp,';');
        char* name=0;
        if (!semi)
          {
            fprintf(stderr,"doxml: expand_entity_str(): no ; after &\n");
            if (res)
              free(res);
            return 0;
          }
        
        {
          int namelen=semi-amp-1;
          name=(char*)malloc(namelen+1);
          if (!name)
            {
              fprintf(stderr,"doxml: expand_entity_str(): can't malloc name\n");
              if (res)
                free(res);
              return 0;
            }
          
          strncpy(name,amp+1,namelen);
          name[namelen]=0;
        }
        
        {
          int must_free_after;
          const char* value
            =get_entity_value_expanded(name,
                                       doctype,
                                       &must_free_after);
          free(name);
          if (!value)
            {
              fprintf(stderr,"doxml: expand_entity_str(): can't malloc name\n");
              if (res)
                free(res);
              return 0;
            }

          {
            char* new_res=append_str(res,&reslen,value,-1);
            if (!new_res)
              {
                fprintf(stderr,
                        "doxml: expand_entity_str(): can't append value\n");
                if (res)
                  free(res);
                if (value && must_free_after)
                  free((char*)value);
                return 0;
              }
            res=new_res;
          }
        }

        entity_str=semi+1;
      }
    }

  if (!res)
    return 0;

  {
    char* new_res=append_str(res,&reslen,entity_str,-1);
    if (!new_res)
      {
        fprintf(stderr,
                "doxml: expand_entity_str(): can't append last bit of string\n");
        if (res)
          free(res);
        return 0;
      }
    res=new_res;
  }
}

static const char*
parse_quoted(doxml_context* context,
             const doxml_doctype* doctype,
             int normalize,
             int is_not_cdata,
             int expand_entities)
{
  char quote=0;
  int ch=doxml_getc(context);
  switch (ch)
    {
    case EOF:
      fprintf(stderr,"doxml: parse_quoted(): EOF before value\n");
      break;

    case '"':
    case '\'':
      quote=ch;
      break;

    default:
      fprintf(stderr,"doxml: parse_quoted(): bad char '%c' as quote on value\n",ch);
      break;
    }
  
  if (!quote)
    return 0;
  
  if (doxml_buff_init(context)<0)
    {
      fprintf(stderr,"doxml: parse_quoted(): buff_init() failed\n");
      return 0;
    }
  
  while ((ch=doxml_getc(context))!=quote)
    {
      if (ch==EOF)
        {
          fprintf(stderr,"doxml: parse_quoted(): EOF in value\n");
          return 0;
        }

      if ((ch=='&') && expand_entities)
        {
          int must_free_after,expandable;
          const char* entity_str=parse_entity(context,doctype,0,0,
                                              &expandable,
                                              &must_free_after);
          if (!entity_str)
            {
              fprintf(stderr,"doxml: parse_quoted(): unrecognized entity\n");
              return 0;
            }

          if (expandable)
            {
              int must_free_expanded;
              const char* expanded=expand_entity_str(entity_str,doctype,
                                                     &must_free_expanded);
              if (expanded)
                {
                  if (must_free_after)
                    free((char*)entity_str);
                  entity_str=expanded;
                  must_free_after=must_free_expanded;
                }
            }

          if (doxml_buff_append_str_collapse_space(context,
                                                   entity_str,
                                                   is_not_cdata)<0)
            {
              fprintf(stderr,"doxml: parse_quoted(): can't append entity-value \"%s\" to buffer\n",
                      entity_str);
              if (must_free_after)
                free((char*)entity_str); /* cast away const */
              return 0;
            }
          if (must_free_after)
            free((char*)entity_str); /* cast away const */
          continue;
        }

      if (normalize && (isspace(ch)))
        ch=' ';

      if (doxml_buff_append_collapse_space(context,ch,is_not_cdata)<0)
        {
          fprintf(stderr,"doxml: parse_quoted(): buff_append() failed\n");
          return 0;
        }
    }
  
  if (ch!=quote) /* can't happen */
    {
      fprintf(stderr,"doxml: parse_quoted(): ended loop with no quote.  Can't happen.\n");
      return 0;
    }
  
  {
    const char* res=doxml_buff_extract(context);
    if (!res)
      fprintf(stderr,"doxml: parse_quoted(): can't extract value\n");
    return res;
  }
}

static doxml_atom*
parse_text(doxml_context* context,
           doxml_atom* parent,
           int had_leading_whitespace,
           const doxml_doctype* doctype)
{
  const doxml_markupdecl* decl=0;
  const char* notation_prefix=0;
  int ch;

#ifdef DEBUG
  fprintf(stderr,"doxml: parse_text()\n");
#endif

  if (doxml_buff_init(context)<0)
    {
      fprintf(stderr,"doxml: parse_text(): can't initialize buffer\n");
      return 0;
    }

  if (had_leading_whitespace)
    {
      if (doxml_buff_append(context,' ')<0)
        {
          fprintf(stderr,"doxml: parse_text(): can't append leading whitespace\n");
          return 0;
        }
    }

 non_whitespace:
  while (!isspace(ch=doxml_getc(context)))
    {
      if (ch==EOF)
        goto extract;

      if (ch=='<')
        {
          if (doxml_ungetc(ch,context)<0)
            {
              fprintf(stderr,"doxml: parse_text(): encountered <, but ungetc failed\n");
              return 0;
            }
          goto extract;
        }

      if (ch=='&')
        goto entity;

      if (doxml_buff_append(context,ch)<0)
        {
          fprintf(stderr,"doxml: parse_text(): can't append character to buffer\n");
          return 0;
        }
    }

 whitespace:
  while(isspace(ch=doxml_getc(context)))
    ;

  if (doxml_buff_append(context,' ')<0)
    {
      fprintf(stderr,"doxml: parse_text(): can't append space to buffer\n");
      return 0;
    }

  if (ch==EOF)
    goto extract;

  doxml_ungetc(ch,context);
  goto non_whitespace;

 entity:
  {
    int must_free_after,expandable;
    const char* entity_str=parse_entity(context,doctype,&decl,&notation_prefix,
                                        &expandable,
                                        &must_free_after);
    if (decl)
      {
        if (entity_str)
          {
            fprintf(stderr,"doxml: parse_text(): unparsed entities are forbidden in text\n");
            if (must_free_after)
              free((char*)entity_str); /* cast away const */
            return 0;
          }
        
        goto extract;
      }

    if (!entity_str)
      {
        fprintf(stderr,"doxml: parse_text(): unrecognized entity\n");
        return 0;
      }
    else
      {
        if (expandable)
          {
            int must_free_expanded;
            const char* expanded=expand_entity_str(entity_str,doctype,
                                                   &must_free_expanded);
            if (expanded)
              {
                if (must_free_after)
                  free((char*)entity_str);
                entity_str=expanded;
                must_free_after=must_free_expanded;
              }
          }
        
        if (doxml_buff_append_str(context,entity_str)<0)
          {
            fprintf(stderr,"doxml: parse_text(): can't append entity-value \"%s\" to buffer\n",
                    entity_str);
            if (must_free_after)
              free((char*)entity_str); /* cast away const */
            return 0;
          }
        if (must_free_after)
          free((char*)entity_str); /* cast away const */
        goto non_whitespace;
      }
  }

 extract:
  {
    doxml_atom* res=0;
    doxml_atom* external_entity_res=0;
    const char* text=doxml_buff_extract(context);
    if (!text)
      {
        fprintf(stderr,"doxml: parse_text(): can't extract text from buffer\n");
        return 0;
      }
    
    res=new_text_atom(parent,text);
    if (!res)
      {
        fprintf(stderr,"doxml: parse_text(): can't create text atom\n");
        return 0;
      }

    if (decl)
      {
        external_entity_res=new_external_entity_atom(parent,decl);
        if (!external_entity_res)
          {
            fprintf(stderr,"doxml: parse_text(): can't create external entity atom\n");
            doxml_delete_atom(res);
            return 0;
          }
      }

    return res;
  }
}

static doxml_atom*
parse_cdata(doxml_context* context,
            doxml_atom* parent)
{
  int ch;

  if (doxml_buff_init(context)<0)
    {
      fprintf(stderr,"doxml: parse_cdata(): can't initialize buffer\n");
      return 0;
    }

  while ((ch=doxml_getc(context))!=EOF)
    {
      if (ch==']')
        {
          ch=doxml_getc(context);
          if (ch==EOF)
            break;
          if (ch==']')
            {
              ch=doxml_getc(context);
              if (ch==EOF)
                break;
              if (ch=='>')
                break;
              if (doxml_buff_append_str(context,"]]")<0)
                {
                  fprintf(stderr,"doxml: parse_cdata(): can't append ]] to buff\n");
                  return 0;
                }
              if (doxml_buff_append(context,ch)<0)
                {
                  fprintf(stderr,"doxml: parse_cdata(): can't append ch to buff\n");
                  return 0;
                }
            }
          else
            {
              if (doxml_buff_append(context,']')<0)
                {
                  fprintf(stderr,"doxml: parse_cdata(): can't append ] to buff\n");
                  return 0;
                }
              if (doxml_buff_append(context,ch)<0)
                {
                  fprintf(stderr,"doxml: parse_cdata(): can't append ch to buff\n");
                  return 0;
                }
            }
        }
      else
        {
          if (doxml_buff_append(context,ch)<0)
            {
              fprintf(stderr,"doxml: parse_cdata(): can't append ch to buff\n");
              return 0;
            }
        }
    }

  if (ch==EOF)
    {
      fprintf(stderr,"doxml: parse_cdata(): found EOF inside a CDATA block\n");
      return 0;
    }

  {
    const char* text=doxml_buff_extract(context);
    if (!text)
      {
        fprintf(stderr,"doxml: parse_cdata(): can't extract text from buffer\n");
        return 0;
      }
    return new_text_atom(parent,text);
  }
}

static doxml_atom*
parse_comment(doxml_context* context,
              int* non_atom)
{
  int ch;
  while ((ch=doxml_getc(context))!=EOF)
    {
      if (ch!='-')
        continue;

      ch=doxml_getc(context);
      if (ch!='-')
        continue;
      
      ch=doxml_getc(context);
      if (ch!='>')
        {
          fprintf(stderr,"doxml: parse_comment(): -- in comment not followed by >\n");
          break;
        }
      
      (*non_atom)=1;
      return 0;
    }

  if (ch==EOF)
    fprintf(stderr,"doxml: parse_comment(): EOF in comment\n");
  (*non_atom)=0;
  return 0;
}

static int
expect(doxml_context* context,
       const char* s)
{
  const char* cur=s;
  if (!s)
    {
      fprintf(stderr,"doxml: expect(): s==NULL\n");
      return -1;
    }

  while (*cur)
    {
      int ch=doxml_getc(context);
      if (ch==EOF)
        {
          fprintf(stderr,"doxml: expect(\"%s\"): EOF\n",s);
          return -1;
        }

      if (ch!=cur[0])
        {
          fprintf(stderr,
                  "doxml: expect(\"%s\"): found '%c' in place of \"%s\"\n",
                  s,ch,cur
                  );
          return -1;
        }

      cur++;
    }

  return 0;
}

static int
expect_poly(doxml_context* context,
            const char* choices[])
{
  int ch,N,numLeft;

  if (!choices)
    {
      fprintf(stderr,"doxml: expect_poly(): choices==NULL\n");
      return -1;
    }

  if (!choices[0])
    {
      fprintf(stderr,"doxml: expect_poly(): choices=={}\n");
      return -1;
    }

#ifdef DEBUG_VERBOSE_EXPECT
  fprintf(stderr,"doxml: expect_poly(): choices=={\n");
#endif
  for (N=0; choices[N]; N++)
#ifdef DEBUG_VERBOSE_EXPECT
    fprintf(stderr,"                                %s\n",choices[N])
#endif
      ;
#ifdef DEBUG_VERBOSE_EXPECT
  fprintf(stderr,"                               }\n");
#endif

  numLeft=N;

  while ((ch=doxml_getc(context))!=EOF)
    {
      int i;
      for (i=0; i<N; i++)
        {
          if (!choices[i])
            continue;

          if (choices[i][0]==ch)
            {
              choices[i]++;
              if (choices[i][0]==0)
                return i;
            }
          else
            {
              choices[i]=0;
              numLeft--;
              if (numLeft<=0)
                {
                  fprintf(stderr,"doxml: %d: expect_poly(): didn't find any legal choices\n",
                          context->line_number);
                  return -1;
                }
            }
        }
    }

  fprintf(stderr,"doxml: expect_poly(): found EOF\n");
  return -1;
}

static int
is_pubid_char(int ch)
{
  if (ch>127)
    return 0;

  if (is_name_char(ch))
    return 1;

  switch (ch)
    {
    case ' ':
    case '\r':
    case '\n':
    case '-':
    case '\'':
    case '(':
    case ')':
    case '+':
    case ',':
    case '/':
    case ':':
    case '=':
    case '?':
    case ';':
    case '!':
    case '*':
    case '#':
    case '@':
    case '$':
    case '_':
    case '%':
      return 1;

    default:
      return 0;
    }
}

static int
parse_external_id(doxml_context* context,
                  doxml_external_id* id,
                  const doxml_doctype* doctype)
{
  int ch;

#ifdef DEBUG
  fprintf(stderr,"doxml: parse_external_id()\n");
#endif

  if (!id)
    {
      fprintf(stderr,"doxml: parse_external_id(): id==NULL\n");
      return -1;
    }

  id->uri_contents=0;

  {
    int which;
    const char* choices[3];
    choices[0]="SYSTEM";
    choices[1]="PUBLIC";
    choices[2]=0;

    which=expect_poly(context,choices);
    if (   (which<0)
        || (which>1) /* can't happen */
        )
      {
        fprintf(stderr,
                "doxml: %d: parse_external_id(): didn't find SYSTEM or PUBLIC\n",
                context->line_number);
        return -1;
      }

    id->is_public=which;
  }

#ifdef DEBUG
  fprintf(stderr,"doxml: parse_external_id(): is %spublic\n",
          (id->is_public ? "" : "not ")
          );
#endif

  ch=skip_ws(context,1);
  if (ch<0)
    {
      fprintf(stderr,"doxml: parse_external_id(): skip_ws() failed\n");
      return -1;
    }

  if (id->is_public)
    {
      doxml_ungetc(ch,context);
      id->pubid=parse_quoted(context,doctype,0,0,0);
      if (!(id->pubid))
        {
          fprintf(stderr,"doxml: parse_external_id(): can't parse public id\n");
          return -1;
        }

      ch=skip_ws(context,1);
      if (ch<0)
        {
          fprintf(stderr,"doxml: parse_external_id(): skip_ws() failed\n");
          return -1;
        }
    }
  else
    id->pubid=0;

  doxml_ungetc(ch,context);
  id->system_uri=parse_quoted(context,doctype,0,0,0);
  if (!(id->system_uri))
    {
      fprintf(stderr,"doxml: parse_external_id(): can't parse system\n");
      if (id->pubid)
        free((char*)(id->pubid));
      return -1;
    }

  return 0;
}

static doxml_external_id*
parse_external_id_alloc(doxml_context* context,
                        const doxml_doctype* doctype)
{
  doxml_external_id* id=(doxml_external_id*)malloc(sizeof(doxml_external_id));
  if (!id)
    {
      fprintf(stderr,"doxml: parse_external_id_alloc(): can't alloc id\n");
      return 0;
    }

  if (parse_external_id(context,id,doctype)<0)
    {
      if (id->pubid) free((char*)(id->pubid));
      if (id->system_uri) free((char*)(id->system_uri));
      free(id);
    }

  return id;
}

static void
delete_markupdecl_element_cp(doxml_markupdecl_element_cp* cp)
{
  if (!cp)
    return;

  if (cp->next)
    delete_markupdecl_element_cp(cp->next);

  switch (cp->type)
    {
    case doxml_markupdecl_element_cp_choice:
      delete_markupdecl_element_cp(cp->data.choice);
      break;
      
    case doxml_markupdecl_element_cp_seq:
      delete_markupdecl_element_cp(cp->data.seq);
      break;
      
    case doxml_markupdecl_element_cp_name:
      if (cp->data.name.prefix)
        free((char*)(cp->data.name.prefix));
      if (cp->data.name.name)
        free((char*)(cp->data.name.name));
      break;

    default:
      fprintf(stderr,"doxml: delete_markupdecl_element_cp(): unknown type\n");
      break;
    }

  free(cp);
}

static void
delete_markupdecl(doxml_markupdecl* decl)
{
  if (!decl)
    return;

  switch (decl->type)
    {
    case doxml_markupdecl_element:
      if (decl->data.element.name.prefix)
        free((char*)(decl->data.element.name.prefix));
      if (decl->data.element.name.name)
        free((char*)(decl->data.element.name.name));
      switch (decl->data.element.type)
        {
        case doxml_markupdecl_element_empty:
        case doxml_markupdecl_element_any:
          break;

        case doxml_markupdecl_element_mixed:
          {
            doxml_name_node* cur;
            doxml_name_node* next;
            for (cur=decl->data.element.data.mixed; cur; cur=next)
              {
                next=cur->next;
                if (cur->name.prefix)
                  free((char*)(cur->name.prefix));
                if (cur->name.name)
                  free((char*)(cur->name.name));
              }
          }
          break;

        case doxml_markupdecl_element_children:
          delete_markupdecl_element_cp(decl->data.element.data.children);
          break;

        default:
          fprintf(stderr,"doxml: delete_markupdecl(): unknown element type\n");
          break;
        }
      break;
      
    case doxml_markupdecl_attlist:
      if (decl->data.attlist.name.prefix)
        free((char*)(decl->data.attlist.name.prefix));
      if (decl->data.attlist.name.name)
        free((char*)(decl->data.attlist.name.name));
      {
        doxml_markupdecl_attdef* cur;
        doxml_markupdecl_attdef* next;
        for (cur=decl->data.attlist.attdefs; cur; cur=next)
          {
            next=cur->next;
            if (cur->name.prefix)
              free((char*)(cur->name.prefix));
            if (cur->name.name)
              free((char*)(cur->name.name));
            if (cur->def.value)
              free((char*)(cur->def.value));
          }
      }
      break;
      
    case doxml_markupdecl_entity:
      if (decl->data.entity.name.name)
        free((char*)(decl->data.entity.name.name));
      if (decl->data.entity.name.prefix)
        free((char*)(decl->data.entity.name.prefix));
      if (decl->data.entity.is_general)
        {
          /*                    no relation */
          if (decl->data.entity.data.general.is_external)
            {
              if (decl->data.entity.data.general.data.external.notation_name.name)
                free((char*)(decl->data.entity.data.general.data.external.notation_name.name));
              if (decl->data.entity.data.general.data.external.notation_name.prefix)
                free((char*)(decl->data.entity.data.general.data.external.notation_name.prefix));
              if (decl->data.entity.data.general.data.external.external_id.pubid)
                free((char*)(decl->data.entity.data.general.data.external.external_id.pubid));
              if (decl->data.entity.data.general.data.external.external_id.system_uri)
                free((char*)(decl->data.entity.data.general.data.external.external_id.system_uri));
            }
          else /* not external */
            {
              if (decl->data.entity.data.general.data.internal)
                free((char*)(decl->data.entity.data.general.data.internal));
            }
        }
      else /* not general */
        {
          if (decl->data.entity.data.param.is_external)
            {
              if (decl->data.entity.data.param.data.external_id.pubid)
                free((char*)(decl->data.entity.data.param.data.external_id.pubid));
              if (decl->data.entity.data.param.data.external_id.system_uri)
                free((char*)(decl->data.entity.data.param.data.external_id.system_uri));
            }
          else /* not external */
            {
              if (decl->data.entity.data.param.data.internal)
                free((char*)(decl->data.entity.data.param.data.internal));
            }
        }
      break;
      
    case doxml_markupdecl_notation:
      if (decl->data.notation.name.prefix)
        free((char*)(decl->data.notation.name.prefix));
      if (decl->data.notation.name.name)
        free((char*)(decl->data.notation.name.name));
      if (decl->data.notation.external_id.pubid)
        free((char*)(decl->data.notation.external_id.pubid));
      if (decl->data.notation.external_id.system_uri)
        free((char*)(decl->data.notation.external_id.system_uri));
      break;
      
    case doxml_markupdecl_pi:
      if (decl->data.pi.name.name)
        free((char*)(decl->data.pi.name.name));
      if (decl->data.pi.name.prefix)
        free((char*)(decl->data.pi.name.prefix));
      if (decl->data.pi.value)
        free((char*)(decl->data.pi.value));
      break;
      
    case doxml_markupdecl_pe_reference:
      if (decl->data.pe_reference.name.prefix)
        free((char*)(decl->data.pe_reference.name.prefix));
      if (decl->data.pe_reference.name.name)
        free((char*)(decl->data.pe_reference.name.name));
      break;
      
    default:
      fprintf(stderr,"doxml: delete_markupdecl(): unknown type\n");
      break;
    }

  free(decl);
}

static void
delete_name_node(doxml_name_node* node)
{
  if (!node)
    return;

  if (node->name.prefix)
    free((char*)(node->name.prefix));
  if (node->name.name)
    free((char*)(node->name.name));

  if (node->next)
    delete_name_node(node->next);

  free(node);
}

static doxml_name_node*
parse_markupdecl_name_node(doxml_context* context,
                           int* err,
                           int allow_nmtokens,
                           int dont_require_initial_vbar
                           )
{
  int ch;

  if (!err)
    {
      fprintf(stderr,"doxml: parse_markupdecl_name_node(): err is NULL\n");
      return 0;
    }

  (*err)=-1;

  if (!context)
    {
      fprintf(stderr,"doxml: parse_markupdecl_name_node(): context is NULL\n");
      return 0;
    }

  ch=skip_ws(context,0);
  switch (ch)
    {
    case EOF:
      fprintf(stderr,"doxml: parse_markupdecl_name_node(): EOF where expected '|' or ')'\n");
      return 0;

    case '|':
    vbar:
      {
        const char* name;
        doxml_name_node* res;

        ch=skip_ws(context,0);
        if (ch==EOF)
          {
            fprintf(stderr,
                    "doxml: parse_markupdecl_name_node(): EOF after '|'\n");
            return 0;
          }

        doxml_ungetc(ch,context);

        name=parse_name_or_nmtoken(context,!allow_nmtokens);
        if (!name)
          {
            fprintf(stderr,"doxml: parse_markupdecl_name_node(): parse_name() failed\n");
            return 0;
          }

        res=(doxml_name_node*)malloc(sizeof(doxml_name_node));
        if (!res)
          {
            fprintf(stderr,"doxml: parse_markupdecl_name_node(): can't malloc res\n");
            free((char*)name);
            return 0;
          }

        res->name.prefix=0;
        res->name.name=name;
        res->prev=0;

        {
          int tmp_err=0;
          res->next=parse_markupdecl_name_node(context,&tmp_err,
                                               allow_nmtokens,0);

          if (tmp_err)
            {
              fprintf(stderr,"doxml: parse_markupdecl_name_node(): recursive call failed\n");
              free((char*)name);
              free(res);
              return 0;
            }
        }

        if (res->next)
          res->next->prev=res;
        (*err)=0;
        return res;
      }
      break;

    case ')': /* end of list */
      (*err)=0;
      return 0;

    default:
      if (dont_require_initial_vbar)
        {
          doxml_ungetc(ch,context);
          goto vbar;
        }

      fprintf(stderr,"doxml: parse_markupdecl_name_node(): bad char '%c' where expected '|' or ')'\n",
              (char)ch);
      return 0;
    }

  fprintf(stderr,"doxml: parse_markupdecl_name_node(): unexpectedly reached end of function\n");
  (*err)=-1;
  return 0;
}


static doxml_name_node*
parse_markupdecl_name_node_with_sep(doxml_context* context,
                                    int* err,
                                    int allow_nmtokens,
                                    int dont_require_initial_sep,
                                    const char* separator,
                                    char terminator
                                    )
{
  int ch;

  if (!err)
    {
      fprintf(stderr,"doxml: parse_markupdecl_name_node_with_sep(): err is NULL\n");
      return 0;
    }

  (*err)=-1;

  if (!context)
    {
      fprintf(stderr,"doxml: parse_markupdecl_name_node_with_sep(): context is NULL\n");
      return 0;
    }

  ch=skip_ws(context,1);
  switch (ch)
    {
    case EOF:
      fprintf(stderr,"doxml: parse_markupdecl_name_node_with_sep(): EOF before separator\n");
      return 0;

    case -2:
      fprintf(stderr,"doxml: parse_markupdecl_name_node_with_sep(): no space before separator\n");
      return 0;
    }

  if (ch==terminator)
    {
      (*err)=0;
      return 0;
    }

  doxml_ungetc(ch,context);

  if (expect(context,separator)<0)
    {
      fprintf(stderr,"doxml: parse_markupdecl_name_node_with_sep(): no separator found\n");
      return 0;
    }

  {
    const char* name;
    doxml_name_node* res;

    ch=skip_ws(context,1);
    switch (ch)
      {
      case EOF:
        fprintf(stderr,
                "doxml: parse_markupdecl_name_node_with_sep(): EOF after separator\n");
        return 0;

      case -2:
        fprintf(stderr,
                "doxml: parse_markupdecl_name_node_with_sep(): no space after separator'\n");
        return 0;
      }

    doxml_ungetc(ch,context);

    name=parse_name_or_nmtoken(context,!allow_nmtokens);
    if (!name)
      {
        fprintf(stderr,"doxml: parse_markupdecl_name_node_with_sep(): parse_name() failed\n");
        return 0;
      }

    res=(doxml_name_node*)malloc(sizeof(doxml_name_node));
    if (!res)
      {
        fprintf(stderr,"doxml: parse_markupdecl_name_node_with_sep(): can't malloc res\n");
        free((char*)name);
        return 0;
      }

    res->name.prefix=0;
    res->name.name=name;
    res->prev=0;

    {
      int tmp_err=0;
      res->next=parse_markupdecl_name_node(context,&tmp_err,
                                           allow_nmtokens,0);

      if (tmp_err)
        {
          fprintf(stderr,"doxml: parse_markupdecl_name_node_with_sep(): recursive call failed\n");
          free((char*)name);
          free(res);
          return 0;
        }
    }

    if (res->next)
      res->next->prev=res;
    (*err)=0;
    return res;
  }

  fprintf(stderr,"doxml: parse_markupdecl_name_node_with_sep(): unexpectedly reached end of function\n");
  (*err)=-1;
  return 0;
}

static doxml_markupdecl_element_cp*
parse_markupdecl_element_cp(doxml_context* context,
                            int already_got_paren)
{
  doxml_markupdecl_element_cp* res=0;

  int ch;

  if (!context)
    {
      fprintf(stderr,"doxml: parse_markupdecl_element_cp(): context is NULL\n");
      return 0;
    }

  res=(doxml_markupdecl_element_cp*)malloc(sizeof(doxml_markupdecl_element_cp));
  if (!res)
    {
      fprintf(stderr,"doxml: parse_markupdecl_element_cp(): can't malloc res\n");
      return 0;
    }

  res->next=res->prev=0;
  res->type=doxml_markupdecl_element_cp_invalid;

  if (!already_got_paren)
    {
      ch=skip_ws(context,0);
      switch (ch)
        {
        case EOF:
          fprintf(stderr,"doxml: parse_markupdecl_element_cp(): EOF at start of a cp\n");
          free(res);
          return 0;

        case '(':
          break;

        default:
          doxml_ungetc(ch,context);
          res->type=doxml_markupdecl_element_cp_name;
          {
            if (parse_name_in_dtd(context,&(res->data.name))<0)
              {
                fprintf(stderr,
                        "doxml: parse_markupdecl_element_cp(): parse_name() failed\n");
                free(res);
                return 0;
              }
          }

          goto multiplicity;
        }
    }

  /* OK, it's either a choice or a seq, and we've skipped the ( */
  {
    doxml_markupdecl_element_cp* first=parse_markupdecl_element_cp(context,0);
    doxml_markupdecl_element_cp* last=first;
    if (!first)
      {
        fprintf(stderr,"doxml: parse_markupdecl_element_cp(): first recursive call failed\n");
        free(res);
        return 0;
      }

  next_cp:
    ch=skip_ws(context,0);
    {
      doxml_markupdecl_element_cp_type sep_type
        =doxml_markupdecl_element_cp_invalid;
      switch (ch)
        {
        case EOF:
          fprintf(stderr,"doxml: parse_markupdecl_element_cp(): EOF instead of separator");
          break;

        case ')':
          if (res->type==doxml_markupdecl_element_cp_invalid)
            {
              /* This is weird.  The spec permits both choice and
               *  seq to accept 1-item lists, such as '(' cp ')'.  This
               *  is both pointless (who needs the extra parens?) and
               *  ambiguous (is it a choice or a seq?).  Fortunately,
               *  it doesn't matter which, since they're equivalent in
               *  this case, so I'm just going to declare it to be
               *  a choice.
               */
              res->type=doxml_markupdecl_element_cp_choice;
              res->data.choice=first;
            }
          goto multiplicity;

        case '|':
          sep_type=doxml_markupdecl_element_cp_choice;
          break;

        case ',':
          sep_type=doxml_markupdecl_element_cp_seq;
          break;

        default:
          fprintf(stderr,"doxml: parse_markupdecl_element_cp(): bad char '%c' instead of separator\n",(char)ch);
          break;
        }
      
      if (sep_type==doxml_markupdecl_element_cp_invalid)
        {
          delete_markupdecl_element_cp(first);
          free(res);
          return 0;
        }
      
      if (res->type==doxml_markupdecl_element_cp_invalid)
        {
          res->type=sep_type;
          if (sep_type==doxml_markupdecl_element_cp_choice)
            res->data.choice=first;
          else
            res->data.seq=first;
        }
      else
        {
          if (res->type!=sep_type)
            {
              fprintf(stderr,"doxml: parse_markupdecl_element_cp(): mismatched separator\n");
              delete_markupdecl_element_cp(res);
              return 0;
            }
        }
      
      {
        doxml_markupdecl_element_cp* next
          =parse_markupdecl_element_cp(context,0);
        if (!next)
          {
            fprintf(stderr,"doxml: parse_markupdecl_element_cp(): next recursive call failed\n");
            delete_markupdecl_element_cp(res);
          }

        last->next=next;
        next->prev=last;
        last=next;
      }
    }

    goto next_cp;
  }

 multiplicity:
  ch=doxml_getc(context);
  switch (ch)
    {
    case EOF:
      fprintf(stderr,"doxml: parse_markupdecl_element_cp(): EOF instead of multiplicity\n");
      delete_markupdecl_element_cp(res);
      return 0;

    case '?':
      res->multiplicity=doxml_multiplicity_1_or_0;
      break;

    case '+':
      res->multiplicity=doxml_multiplicity_1_or_more;
      break;

    case '*':
      res->multiplicity=doxml_multiplicity_0_or_more;
      break;

    default:
      doxml_ungetc(ch,context);
      res->multiplicity=doxml_multiplicity_1;
      break;
    }

  return res;
}

static doxml_markupdecl_attdef*
parse_attdefs(doxml_context* context,
              int* err,
              const doxml_doctype* doctype)
{
  int ch;
  doxml_markupdecl_attdef* res=0;

  if (!err)
    {
      fprintf(stderr,"doxml: parse_attdefs(): err is NULL\n");
      return 0;
    }

  (*err)=-1;

  if (!context)
    {
      fprintf(stderr,"doxml: parse_attdefs(): context is NULL\n");
      return 0;
    }

  ch=skip_ws(context,0);
  if (ch==EOF)
    {
      fprintf(stderr,"doxml: parse_attdefs(): EOF in ATTLIST\n");
      return 0;
    }

  if (ch=='>')
    {
      *err=0;
      return 0;
    }

  doxml_ungetc(ch,context);

  res=(doxml_markupdecl_attdef*)malloc(sizeof(doxml_markupdecl_attdef));
  if (!res)
    {
      fprintf(stderr,"doxml: parse_attdefs(): can't malloc res\n");
      return 0;
    }

  if (parse_name_in_dtd(context,&(res->name))<0)
    {
      fprintf(stderr,"doxml: parse_attdefs(): parse_name_in_dtd() failed\n");
      free(res);
      return 0;
    }

  ch=skip_ws(context,1);
  if (ch<0)
    {
      if (ch==EOF)
        fprintf(stderr,"doxml: parse_attdefs(): EOF before attype\n");
      else
        fprintf(stderr,"doxml: parse_attdefs(): no space before attype\n");
      if (res->name.prefix)
        free((char*)(res->name.prefix));
      free((char*)(res->name.name));
      free(res);
      return 0;
    }

  if (ch=='(')
    {
      int tmp_err;
      res->attype.type=doxml_markupdecl_attdef_attype_enumerated;
      res->attype.enumeration.is_notation=0;
      res->attype.enumeration.names=parse_markupdecl_name_node(context,
                                                               &tmp_err,
                                                               0,1);
      if (tmp_err)
        fprintf(stderr,"doxml: parse_attdefs(): parse_markupdecl_name_node() failed\n");
      else
        {
          if (!(res->attype.enumeration.names))
            {
              fprintf(stderr,"doxml: parse_attdefs(): an enumeration must contain at least 1 name\n");
              tmp_err=-1;
            }
        }

      if (tmp_err)
        {
          if (res->name.prefix)
            free((char*)(res->name.prefix));
          free((char*)(res->name.name));
          free(res);

          return 0;
        }
    }
  else
    {
      const char* choices[]
        ={"CDATA",
            "ID","IDREF","IDREFS",
            "ENTITY","ENTITIES",
            "NMTOKEN","NMTOKENS",
            "NOTATION",
            0};
      int which;

      doxml_ungetc(ch,context);
      which=expect_poly(context,choices);
      if ((which<0) || (which>8))
        {
          fprintf(stderr,"doxml: parse_attdefs(): expect_poly() did not find any legal choice\n");
          if (res->name.prefix)
            free((char*)(res->name.prefix));
          free((char*)(res->name.name));
          free(res);
          return 0;
        }

      if (which<8)
        {
          res->attype.type=doxml_markupdecl_attdef_attype_cdata+which;
          res->attype.enumeration.names=0;
          res->attype.enumeration.is_notation=0;
        }
      else
        {
          int is_bad=1;
          ch=skip_ws(context,1);
          switch (ch)
            {
            case EOF:
              fprintf(stderr,"doxml: parse_attdefs(): EOF after NOTATION\n");
              break;

            case -2:
              fprintf(stderr,"doxml: parse_attdefs(): no space after NOTATION\n");
              break;

            case '(':
              is_bad=0;
              break;

            default:
              fprintf(stderr,"doxml: parse_attdefs(): bad char '%c' after NOTATION\n");
              break;
            }

          if (is_bad)
            {
              if (res->name.prefix)
                free((char*)(res->name.prefix));
              free((char*)(res->name.name));
              free(res);
              return 0;
            }
          
          {
            int tmp_err;
            res->attype.type=doxml_markupdecl_attdef_attype_enumerated;
            res->attype.enumeration.is_notation=1;
            res->attype.enumeration.names=parse_markupdecl_name_node(context,
                                                                     &tmp_err,
                                                                     1,1);
            if (tmp_err)
              fprintf(stderr,"doxml: parse_attdefs(): parse_markupdecl_name_node() failed\n");
            else
              {
                if (!(res->attype.enumeration.names))
                  {
                    fprintf(stderr,"doxml: parse_attdefs(): an enumeration must contain at least 1 name\n");
                    tmp_err=-1;
                  }
              }
            
            if (tmp_err)
              {
                if (res->name.prefix)
                  free((char*)(res->name.prefix));
                free((char*)(res->name.name));
                free(res);

                return 0;
              }
          }
        }
    }

  ch=skip_ws(context,1);
  if (ch<0)
    {
      if (ch==EOF)
        fprintf(stderr,"doxml: parse_attdefs(): EOF before attype\n");
      else
        fprintf(stderr,"doxml: parse_attdefs(): no space before attype\n");
      if (res->attype.enumeration.names)
        delete_name_node(res->attype.enumeration.names);
      if (res->name.prefix)
        free((char*)(res->name.prefix));
      free((char*)(res->name.name));
      free(res);
      return 0;
    }

  if (ch=='#')
    {
      const char* choices[]
        ={"REQUIRED","IMPLIED","FIXED",0};
      int which=expect_poly(context,choices);
      if ((which<0) || (which>3))
        {
          fprintf(stderr,"doxml: parse_attdefs(): expect_poly() found no legal choice\n");
          if (res->attype.enumeration.names)
            delete_name_node(res->attype.enumeration.names);
          if (res->name.prefix)
            free((char*)(res->name.prefix));
          free((char*)(res->name.name));
          free(res);
          return 0;
        }
      res->def.type=doxml_markupdecl_attdef_default_required+which;
    }
  else
    {
      doxml_ungetc(ch,context);
      res->def.type=doxml_markupdecl_attdef_default_plain;
    }

  switch (res->def.type)
    {
    case doxml_markupdecl_attdef_default_fixed:
      {
        int bad=0;
        ch=skip_ws(context,1);
        switch (ch)
          {
          case EOF:
            fprintf(stderr,"doxml: parse_attdefs(): EOF before value\n");
            bad=1;
            break;

          case -2:
            fprintf(stderr,
                    "doxml: parse_attdefs(): no space before value\n");
            bad=1;
            break;

          default:
            doxml_ungetc(ch,context);
            break;
          }
        if (bad)
          {
            if (res->attype.enumeration.names)
              delete_name_node(res->attype.enumeration.names);
            if (res->name.prefix)
              free((char*)(res->name.prefix));
            free((char*)(res->name.name));
            free(res);
            return 0;
          }
      }
      /* falling through on purpose */
    case doxml_markupdecl_attdef_default_plain:
      {
        res->def.value=parse_quoted(context,doctype,0,0,1);
        if (!(res->def.value))
          {
            fprintf(stderr,"doxml: parse_attdefs(): cannot extract value\n");
            if (res->attype.enumeration.names)
              delete_name_node(res->attype.enumeration.names);
            if (res->name.prefix)
              free((char*)(res->name.prefix));
            free((char*)(res->name.name));
            free(res);
            return 0;
          }
      }
      break;

    default:
      res->def.value=0;
      break;
    }

  {
    int tmp_err;
    res->next=parse_attdefs(context,&tmp_err,doctype);
    if (tmp_err)
      {
        fprintf(stderr,"doxml: parse_attdefs(): recursive call failed\n");
        if (res->def.value)
          free((char*)(res->def.value));
        if (res->attype.enumeration.names)
          delete_name_node(res->attype.enumeration.names);
        if (res->name.prefix)
          free((char*)(res->name.prefix));
        free((char*)(res->name.name));
        free(res);
        return 0;
      }
    if (res->next)
      res->next->prev=res;
  }

  *err=0;
  return res;
}

static int
parse_markupdecl_notation(doxml_context* context,
                          const doxml_doctype* doctype,
                          doxml_markupdecl* res)
{
  {
    int ch=skip_ws(context,1);
    
    switch (ch)
      {
      case EOF:
        fprintf(stderr,
                "doxml: parse_markupdecl_notation(): EOF after NOTATION\n");
        return -1;

      case -2:
        fprintf(stderr,
                "doxml: parse_markupdecl_notation(): no space after NOTATION\n");
        return -1;

      default:
        doxml_ungetc(ch,context);
        break;
      }
  }

  if (parse_name_in_dtd(context,&(res->data.notation.name))<0)
    {
      fprintf(stderr,"doxml: parse_markupdecl_notation(): can't parse name\n");
      return -1;
    }

  {
    int ch=skip_ws(context,1);
    
    switch (ch)
      {
      case EOF:
        fprintf(stderr,
                "doxml: parse_markupdecl_notation(): EOF after name\n");
        return -1;

      case -2:
        fprintf(stderr,
                "doxml: parse_markupdecl_notation(): no space after name\n");
        return -1;

      default:
        doxml_ungetc(ch,context);
        break;
      }
  }

  if (parse_external_id(context,&(res->data.notation.external_id),doctype)<0)
    {
      if (res->data.notation.name.name)
        free((char*)(res->data.notation.name.name));
      if (res->data.notation.name.prefix)
        free((char*)(res->data.notation.name.prefix));
      fprintf(stderr,
              "doxml: %d: parse_markupdecl_notation(): can't parse external-id\n",
              context->line_number);
      return -1;
    }


  {
    int ch=skip_ws(context,0);
    
    switch (ch)
      {
      case EOF:
        fprintf(stderr,
                "doxml: parse_markupdecl_notation(): EOF after name\n");
        return -1;

      case '>':
        break;

      default:
        fprintf(stderr,
                "doxml: %d: parse_markupdecl_notation(): illegal character %c after external-id\n",
                context->line_number,
                (char)ch
                );
        return -1;
      }
  }

  return 0;
}

static int
parse_markupdecl(doxml_context* context,
                 doxml_markupdecl** first,
                 doxml_markupdecl** last,
                 const doxml_doctype* doctype)
{
  doxml_markupdecl* res;
  int ch;

  if (!(first && last))
    {
      fprintf(stderr,"doxml: parse_markupdecl(): first or last is NULL\n");
      return -1;
    }

  if (!context)
    {
      fprintf(stderr,"doxml: parse_markupdecl(): context is NULL\n");
      return -1;
    }

  ch=skip_ws(context,0);
  switch (ch)
    {
    case EOF:
      fprintf(stderr,"doxml: parse_markupdecl(): EOF before markupdecl\n");
      return -1;

    case ']':
      return 1;

    case '%':
      fprintf(stderr,"doxml: parse_markupdecl(): PEreferences not yet implemented\n");
      return -1;

    case '<':
      {
        ch=doxml_getc(context);
        switch (ch)
          {
          case '!':
            {
              doxml_markupdecl_type type;
              {
                const char* choices[]
                  ={"ELEMENT","ATTLIST","ENTITY","NOTATION","--",0};
                int which=expect_poly(context,choices);
              
                if ((which<0) || (which>4))
                  {
                    fprintf(stderr,"doxml: parse_markupdecl(): expect_poly() did not find any legal choice\n");
                    return -1;
                  }

                if (which==4)
                  {
                    int non_atom;
                    parse_comment(context,&non_atom);
                    if (!non_atom)
                      {
                        fprintf(stderr,"doxml: parse_markupdecl(): parse_comment() failed\n");
                        return -1;
                      }
                    return 0;
                  }

                type=doxml_markupdecl_element+which;
              }

              res=(doxml_markupdecl*)malloc(sizeof(doxml_markupdecl));
              if (!res)
                {
                  fprintf(stderr,
                          "doxml: parse_markupdecl(): could not malloc res\n");
                  return -1;
                }
              
              res->type=type;

              switch (type)
                {
                case doxml_markupdecl_element:
                  ch=skip_ws(context,1);
                  switch (ch)
                    {
                    case EOF:
                      fprintf(stderr,
                              "doxml: parse_markupdecl(): EOF between ELEMENT and name\n");
                      free(res);
                      return -1;

                    case -2:
                      fprintf(stderr,
                              "doxml: parse_markupdecl(): no space after ELEMENT\n");
                      free(res);
                      return -1;

                    default:
                      doxml_ungetc(ch,context);
                      break;
                    }

                  if (parse_name_in_dtd(context,&(res->data.element.name))<0)
                      {
                        fprintf(stderr,
                                "doxml: parse_markupdecl(): parse_name_in_dtd() failed\n");
                        free(res);
                        return -1;
                      }

                  ch=skip_ws(context,1);
                  switch (ch)
                    {
                    case EOF:
                      fprintf(stderr,
                              "doxml: parse_markupdecl(): EOF after ELEMENT name\n");
                      if (res->data.element.name.prefix)
                        free((char*)(res->data.element.name.prefix));
                      free((char*)(res->data.element.name.name));
                      free(res);
                      return -1;

                    case -2:
                      fprintf(stderr,
                              "doxml: parse_markupdecl(): no space after ELEMENT name\n");
                      if (res->data.element.name.prefix)
                        free((char*)(res->data.element.name.prefix));
                      free((char*)(res->data.element.name.name));
                      free(res);
                      return -1;

                    case 'E':
                      if (expect(context,"MPTY")<0)
                        {
                          fprintf(stderr,
                                  "doxml: parse_markupdecl(): unknown contentspec\n");
                          if (res->data.element.name.prefix)
                            free((char*)(res->data.element.name.prefix));
                          free((char*)(res->data.element.name.name));
                          free(res);
                          return -1;

                        }
                      res->data.element.type=doxml_markupdecl_element_empty;
                      break;

                    case 'A':
                      if (expect(context,"NY")<0)
                        {
                          fprintf(stderr,
                                  "doxml: parse_markupdecl(): unknown contentspec\n");
                          if (res->data.element.name.prefix)
                            free((char*)(res->data.element.name.prefix));
                          free((char*)(res->data.element.name.name));
                          free(res);
                          return -1;
                        }
                      res->data.element.type=doxml_markupdecl_element_any;
                      break;

                    case '(':
                      ch=skip_ws(context,0);
                      if (ch=='#')
                        {
                          if (expect(context,"PCDATA")<0)
                            {
                              fprintf(stderr,
                                      "doxml: parse_markupdecl(): # not followed by PCDATA\n");
                              if (res->data.element.name.prefix)
                                free((char*)(res->data.element.name.prefix));
                              free((char*)(res->data.element.name.name));
                              free(res);
                              return -1;
                            }
                          res->data.element.type
                            =doxml_markupdecl_element_mixed;
                          {
                            int err;
                            res->data.element.data.mixed
                              =parse_markupdecl_name_node(context,&err,0,0);
                            if (err)
                              {
                                fprintf(stderr,
                                        "doxml: parse_markupdecl(): parse_markupdecl_name_node() failed\n");
                                if (res->data.element.name.prefix)
                                  free((char*)(res->data.element.name.prefix));
                                free((char*)(res->data.element.name.name));
                                free(res);
                                return -1;
                              }

                            if (res->data.element.data.mixed)
                              {
                                ch=doxml_getc(context);
                                if (ch!='*')
                                  {
                                    fprintf(stderr,"doxml: parse_markupdecl(): true mixed followed by '%c' instead of '*'\n",
                                            (char)ch);
                                    delete_name_node(res->data.element.data.mixed);
                                    if (res->data.element.name.prefix)
                                      free((char*)(res->data.element.name.prefix));
                                    free((char*)(res->data.element.name.name));
                                    free(res);
                                    return -1;
                                  }
                              }
                          }
                        }
                      else
                        {
                          res->data.element.type
                            =doxml_markupdecl_element_children;
                          doxml_ungetc(ch,context);
                          res->data.element.data.children
                            =parse_markupdecl_element_cp(context,1);
                          if (!(res->data.element.data.children))
                            {
                              fprintf(stderr,
                                      "doxml: parse_markupdecl(): parse_markupdecl_element_cp() failed\n");
                              if (res->data.element.name.prefix)
                                free((char*)(res->data.element.name.prefix));
                              free((char*)(res->data.element.name.name));
                              free(res);
                              return -1;
                            }
                        }
                      break;
                    }

                  ch=skip_ws(context,0);
                  if (ch!='>')
                    {
                      if (ch==EOF)
                        fprintf(stderr,"doxml: parse_markupdecl(): EOF in ELEMENT\n");
                      else
                        fprintf(stderr,"doxml: %d: parse_markupdecl(): bad char '%c' after ELEMENT\n",
                                context->line_number);
                      if (res->data.element.name.prefix)
                        free((char*)(res->data.element.name.prefix));
                      free((char*)(res->data.element.name.name));
                      free(res);
                      return -1;
                    }
                  break;
                  
                case doxml_markupdecl_attlist:
                  ch=skip_ws(context,1);
                  if (ch<0)
                    {
                      if (ch==EOF)
                        fprintf(stderr,"doxml: parse_markupdecl(): EOF after ATTLIST\n");
                      else
                        fprintf(stderr,"doxml: parse_markupdecl(): no space after ATTLIST\n");
                      free(res);
                      return -1;
                    }

                  doxml_ungetc(ch,context);
                  if (parse_name_in_dtd(context,
                                        &(res->data.attlist.name)
                                        )<0)
                    {
                      fprintf(stderr,"doxml: parse_markupdecl(): parse_name_in_dtd() failed\n");
                      free(res);
                      return -1;
                    }
                  {
                    int err;
                    res->data.attlist.attdefs
                      =parse_attdefs(context,&err,doctype);
                    if (err)
                      {
                        fprintf(stderr,"doxml: parse_markupdecl(): parse_attdefs() failed\n");
                        if (res->data.attlist.name.prefix)
                          free((char*)(res->data.attlist.name.prefix));
                        free((char*)(res->data.attlist.name.name));
                        free(res);
                        return -1;
                      }
                  }
                  break;
                  
                case doxml_markupdecl_entity:
                  ch=skip_ws(context,1);
                  if (ch<0)
                    {
                      if (ch==EOF)
                        fprintf(stderr,"doxml: parse_markupdecl(): EOF after ENTITY\n");
                      else
                        fprintf(stderr,"doxml: parse_markupdecl(): no space after ENTITY\n");
                      free(res);
                      return -1;
                    }

                  res->data.entity.is_general=(ch!='%');
                  if (!(res->data.entity.is_general))
                    {
                      ch=skip_ws(context,1);
                      if (ch<0)
                        {
                          if (ch==EOF)
                            fprintf(stderr,"doxml: parse_markupdecl(): EOF after ENTITY %\n");
                          else
                            fprintf(stderr,"doxml: parse_markupdecl(): no space after ENTITY %\n");
                          free(res);
                          return -1;
                        }
                    }
                  doxml_ungetc(ch,context);
                  if (parse_name_in_dtd(context,
                                        &(res->data.entity.name)
                                        )
                      <0)
                    {
                      fprintf(stderr,"doxml: parse_markupdecl(): parse_name_in_dtd() failed reading entity name\n");
                      free(res);
                      return -1;
                    }

#ifdef DEBUG
                  fprintf(stderr,"doxml: parse_markupdecl(): entity name: parse_name_in_dtd() found ");
                  if (res->data.entity.name.prefix)
                    fprintf(stderr,"%s:",res->data.entity.name.prefix);
                  fprintf(stderr,"%s\n",res->data.entity.name.name);
#endif

                  ch=skip_ws(context,1);
                  if (ch<0)
                    {
                      if (ch==EOF)
                        fprintf(stderr,"doxml: parse_markupdecl(): EOF after ENTITY name\n");
                      else
                        fprintf(stderr,"doxml: parse_markupdecl(): no space after ENTITY name\n");
                      free(res);
                      return -1;
                    }

                  doxml_ungetc(ch,context);

                  if (res->data.entity.is_general)
                    {
                      res->data.entity.data.general.is_external
                        =!((ch=='"') || (ch=='\''));
                      res->data.entity.data.general.parsed=0;
                      res->data.entity.data.general.parsing=0;
                      if (res->data.entity.data.general.is_external)
                        {
                          if (parse_external_id(context,
                                                &(res->data.entity.data.general.data.external.external_id),
                                                doctype
                                                )
                              <0
                              )
                            {
                              fprintf(stderr,"doxml: parse_markupdecl(): could not parse ENTITY external-id\n");
                              if (res->data.entity.name.prefix)
                                free((char*)(res->data.entity.name.prefix));
                              free((char*)(res->data.entity.name.name));
                              free(res);
                              return -1;
                            }

                          res->data.entity.data.general.data.external.notation_name.name=0;
                          res->data.entity.data.general.data.external.notation_name.prefix=0;

                          ch=skip_ws(context,0);
                          if (ch==EOF)
                            {
                              fprintf(stderr,"doxml: parse_markupdecl(): EOF when looking for > or NDATA at end of ENTITY\n");
                              if (res->data.entity.data.general.data.external.external_id.pubid)
                                free((char*)(res->data.entity.data.general.data.external.external_id.pubid));
                              if (res->data.entity.data.general.data.external.external_id.system_uri)
                                free((char*)(res->data.entity.data.general.data.external.external_id.system_uri));
                              if (res->data.entity.name.prefix)
                                free((char*)(res->data.entity.name.prefix));
                              free((char*)(res->data.entity.name.name));
                              free(res);
                              return -1;
                            }

                          if (ch=='N')
                            {
                              if (expect(context,"DATA")<0)
                                {
                                  fprintf(stderr,"doxml: parse_markupdecl(): expect() did not find DATA after N\n");
                                  if (res->data.entity.data.general.data.external.external_id.pubid)
                                    free((char*)(res->data.entity.data.general.data.external.external_id.pubid));
                                  if (res->data.entity.data.general.data.external.external_id.system_uri)
                                    free((char*)(res->data.entity.data.general.data.external.external_id.system_uri));
                                  if (res->data.entity.name.prefix)
                                    free((char*)(res->data.entity.name.prefix));
                                  free((char*)(res->data.entity.name.name));
                                  free(res);
                                  return -1;
                                }

                              ch=skip_ws(context,1);
                              if (ch<0)
                                {
                                  switch (ch)
                                    {
                                    case -2:
                                      fprintf(stderr,"doxml: parse_markupdecl(): no space after NDATA\n");
                                      break;

                                    case EOF:
                                      fprintf(stderr,"doxml: parse_markupdecl(): EOF after NDATA\n");
                                      break;

                                    default:
                                      fprintf(stderr,"doxml: parse_markupdecl(): skip_ws() failed\n");
                                      break;
                                    }

                                  if (res->data.entity.data.general.data.external.external_id.pubid)
                                    free((char*)(res->data.entity.data.general.data.external.external_id.pubid));
                                  if (res->data.entity.data.general.data.external.external_id.system_uri)
                                    free((char*)(res->data.entity.data.general.data.external.external_id.system_uri));
                                  if (res->data.entity.name.prefix)
                                    free((char*)(res->data.entity.name.prefix));
                                  free((char*)(res->data.entity.name.name));
                                  free(res);
                                  return -1;
                                }

                              doxml_ungetc(ch,context);

                              if (parse_name_in_dtd(context,
                                                    &(res->data.entity.data.general.data.external.notation_name)
                                                    )
                                  <0)
                                {
                                  fprintf(stderr,"doxml: parse_markupdecl(): could not parse name after NDATA\n");
                                  if (res->data.entity.data.general.data.external.external_id.pubid)
                                    free((char*)(res->data.entity.data.general.data.external.external_id.pubid));
                                  if (res->data.entity.data.general.data.external.external_id.system_uri)
                                    free((char*)(res->data.entity.data.general.data.external.external_id.system_uri));
                                  if (res->data.entity.name.prefix)
                                    free((char*)(res->data.entity.name.prefix));
                                  free((char*)(res->data.entity.name.name));
                                  free(res);
                                  return -1;
                                }
                            }
                          else
                            doxml_ungetc(ch,context);
                        }
                      else /* not external */
                        {
                          res->data.entity.data.general.data.internal
                            =parse_quoted(context,doctype,0,0,0);
                          if (!(res->data.entity.data.general.data.internal))
                            {
                              if (res->data.entity.name.prefix)
                                free((char*)(res->data.entity.name.prefix));
                              free((char*)(res->data.entity.name.name));
                              free(res);
                              return -1;
                            }
                        }
                    }
                  else /* not general */
                    {
                      res->data.entity.data.param.is_external
                        =!((ch=='"') || (ch=='\''));
                      if (res->data.entity.data.param.is_external)
                        {
                          if (parse_external_id(context,
                                                &(res->data.entity.data.param.data.external_id),
                                                doctype
                                                )
                              <0
                              )
                            {
                              fprintf(stderr,"doxml: parse_markupdecl(): could not parse ENTITY external-id\n");
                              if (res->data.entity.name.prefix)
                                free((char*)(res->data.entity.name.prefix));
                              free((char*)(res->data.entity.name.name));
                              free(res);
                              return -1;
                            }
                        }
                      else /* not external */
                        {
                          res->data.entity.data.param.data.internal
                            =parse_quoted(context,doctype,0,0,0);
                          if (!(res->data.entity.data.param.data.internal))
                            {
                              if (res->data.entity.name.prefix)
                                free((char*)(res->data.entity.name.prefix));
                              free((char*)(res->data.entity.name.name));
                              free(res);
                              return -1;
                            }
                        }
                    }
                  ch=skip_ws(context,0);
                  if (ch==EOF)
                    {
                      fprintf(stderr,"doxml: parse_markupdecl(): EOF when looking for > at end of ENTITY\n");
                      fprintf(stderr,"doxml: parse_markupdecl(): skipping freeing res\n");
                      return -1;
                    }

                  if (ch!='>')
                    {
                      fprintf(stderr,"doxml: %d: parse_markupdecl(): bad char '%c' when looking for > at end of ENTITY\n",
                              context->line_number,
                              ch);
                      fprintf(stderr,"doxml: parse_markupdecl(): skipping freeing res\n");
                      return -1;
                    }
                  break;
                  
                case doxml_markupdecl_notation:
                  {
                    int status=parse_markupdecl_notation(context,
                                                         doctype,
                                                         res);
                    if (status<0)
                      {
                        free(res);
                        return -1;
                      }
                  }
                  break;
                }
            }
            break;

          case '?':
            {
              doxml_atom* pi=parse_pi(context,0);
              if (!pi)
                {
                  fprintf(stderr,
                          "doxml: parse_markupdecl(): parse_pi() failed\n");
                  return -1;
                }

              if (pi->type!=doxml_atom_pi)
                {
                  fprintf(stderr,
                          "doxml: parse_markupdecl(): parse_pi() returned non-PI atom\n");
                  doxml_delete_atom(pi);
                  return -1;
                }

              {
                const char* prefix=pi->data.pi.name.prefix;
                const char* name=pi->data.pi.name.name;
                const char* value=pi->data.pi.value;

                if (!name)
                  {
                    fprintf(stderr,
                            "doxml: parse_markupdecl(): parse_pi() returned NULL name\n");
                    doxml_delete_atom(pi);
                    return -1;
                  }

                /* Make sure that doxml_delete_atom() doesn't free the
                 *  strings.
                 */
                pi->data.pi.name.prefix
                  =pi->data.pi.name.name
                    =pi->data.pi.value
                      =0;
                doxml_delete_atom(pi);

                res=(doxml_markupdecl*)malloc(sizeof(doxml_markupdecl));
                if (!res)
                  {
                    fprintf(stderr,"doxml: parse_markupdecl(): cannot malloc res\n");
                    if (prefix)
                      free((char*)prefix);
                    free((char*)name);
                    if (value)
                      free((char*)value);
                    return -1;
                  }

                res->type=doxml_markupdecl_pi;
                res->data.pi.name.prefix=0;
                res->data.pi.name.name=name;
                res->data.pi.value=value;
              }
            }
          }
      }
      break;

    default:
      fprintf(stderr,
              "doxml: parse_markupdecl(): %d: bad char '%c' at start of markup\n",
              context->line_number,
              (int)ch);
      return -1;
    }

  if (res)
    {
      res->next=0;
      if (res->prev=(*last))
        res->prev->next=res;

      if (!(*first))
        (*first)=res;
      (*last)=res;

      return 0;
    }
  else
    return -1;
}

static doxml_doctype*
parse_doctype(doxml_context* context)
{
  doxml_doctype* res;
  int ch=skip_ws(context,1);
  if (ch==EOF)
    {
      fprintf(stderr,"doxml: parse_doctype(): EOF in whitespace at start of doctype\n");
      return 0;
    }

  if (ch==-2)
    {
      fprintf(stderr,"doxml: parse_doctype(): no whitespace at start of doctype\n");
      return 0;
    }

  doxml_ungetc(ch,context);

  res=(doxml_doctype*)malloc(sizeof(doxml_doctype));
  if (!res)
    {
      fprintf(stderr,"doxml: parse_doctype(): cannot allocate doxml_doctype: out of memory\n");
      return 0;
    }
  res->decls.first=0;
  res->decls.last=0;

  context->doctype=res;

  /* Get the DOCTYPE's name */
  if (parse_name_in_dtd(context,&(res->name))<0)
    {
      fprintf(stderr,"doxml: parse_doctype(): cannot parse name\n");
      free(res);
      return 0;
    }

#ifdef DEBUG
  fprintf(stderr,"doxml: parse_doctype(): name==%s\n",res->name.name);
#endif

  if ((ch=skip_ws(context,0))==EOF)
    {
      fprintf(stderr,"doxml: parse_doctype(): EOF in whitespace after name\n");
      if (res->name.prefix)
        free((char*)res->name.prefix);
      free((char*)(res->name.name));
      free(res);
      return 0;
    }

#ifdef DEBUG
  fprintf(stderr,"doxml: parse_doctype(): after name is %c\n",(char)ch);
#endif

  /* Is there anything after the name? */
  if (ch!='>')
    {
      /* Do we have an ExternalID? */
      if (ch!='[')
        {
          doxml_ungetc(ch,context);

          /* Parse the ExternalID */
          res->external_id
            =(doxml_external_id*)malloc(sizeof(doxml_external_id));
          if (!(res->external_id))
            {
              fprintf(stderr,
                      "doxml: parse_doctype(): cannot malloc external id\n");
              if (res->name.prefix)
                free((char*)res->name.prefix);
              free((char*)(res->name.name));
              free(res);
              return 0;
            }

          if (parse_external_id(context,res->external_id,res)<0)
            {
              fprintf(stderr,
                      "doxml: parse_doctype(): cannot parse external id\n");
              free(res->external_id);
              if (res->name.prefix)
                free((char*)res->name.prefix);
              free((char*)(res->name.name));
              free(res);
              return 0;
            }

          ch=skip_ws(context,0);

          /* After the ExternalID we have to have either declarations or
           *  the end of the DOCTYPE.
           */
          if ((ch!='[') && (ch!='>'))
            {
              fprintf(stderr,"doxml: parse_doctype(): illegal character '%c' after external id\n",ch);
              if (res->external_id->pubid)
                free((char*)(res->external_id->pubid));
              if (res->external_id->system_uri)
                free((char*)(res->external_id->system_uri));
              free(res->external_id);
              if (res->name.prefix)
                free((char*)res->name.prefix);
              free((char*)(res->name.name));
              free(res);
              return 0;
            }
        }

      /* Do we have declarations now? */
      if (ch=='[')
        {
          int markup_res;
          while ((markup_res=parse_markupdecl(context,
                                              &(res->decls.first),
                                              &(res->decls.last),
                                              res
                                              )
                  )==0
                 )
            ;

          if (markup_res<0)
            fprintf(stderr,
                    "doxml: parse_doctype(): parse_markupdecl() failed\n");
          else
            {
              ch=skip_ws(context,0);
              if (ch!='>')
                {
                  if (ch==EOF)
                    fprintf(stderr,
                            "doxml: parse_doctype(): EOF after markup\n");
                  else
                    fprintf(stderr,
                            "doxml: parse_doctype(): bad char '%c' after markup\n",
                            (char)ch);
                  markup_res=-1;
                }
            }

          if (markup_res<0)
            {
              {
                doxml_markupdecl* cur;
                doxml_markupdecl* next;
                for (cur=res->decls.first; cur; cur=next)
                  {
                    next=cur->next;
                    delete_markupdecl(cur);
                  }
              }
              if (res->external_id)
                {
                  if (res->external_id->pubid)
                    free((char*)(res->external_id->pubid));
                  if (res->external_id->system_uri)
                    free((char*)(res->external_id->system_uri));
                  free(res->external_id);
                }
              if (res->name.prefix)
                free((char*)res->name.prefix);
              free((char*)(res->name.name));
              free(res);
              return 0;
            }
        }
    }

  context->doctype=0;
  return res;
}

static doxml_atom*
parse_banger(doxml_context* context,
             doxml_atom* parent,
             doxml_doctype** doctype_out,
             const doxml_doctype* doctype_in,
             int* non_atom)
{
  int ch;

  ch=doxml_getc(context);
  switch (ch)
    {
    case '-':
      ch=doxml_getc(context);
      switch (ch)
        {
        case '-':
          return parse_comment(context,non_atom);
        }
      break;

    case 'D':
      ch=doxml_getc(context);
      switch (ch)
        {
        case 'O':
          ch=doxml_getc(context);
          switch (ch)
            {
            case 'C':
              ch=doxml_getc(context);
              switch (ch)
                {
                case 'T':
                  ch=doxml_getc(context);
                  switch (ch)
                    {
                    case 'Y':
                      ch=doxml_getc(context);
                      switch (ch)
                        {
                        case 'P':
                          ch=doxml_getc(context);
                          switch (ch)
                            {
                            case 'E':
                              if (parent || (!doctype_out))
                                {
                                  fprintf(stderr,"doxml: parse_banger(): found <!DOCTYPE outside of prolog\n");
                                  return 0;
                                }
                              (*doctype_out)=parse_doctype(context);
                              if (!(*(doctype_out)))
                                {
                                  fprintf(stderr,"doxml: parse_doctype() failed\n");
                                  return 0;
                                }
                              *non_atom=1;
                              return 0;
                            }
                          break;
                        }
                      break;
                    }
                  break;
                }
              break;
            }
          break;
        }
      break;

    case 'C':
      ch=doxml_getc(context);
      switch (ch)
        {
        case 'D':
          ch=doxml_getc(context);
          switch (ch)
            {
            case 'A':
              ch=doxml_getc(context);
              switch (ch)
                {
                case 'T':
                  ch=doxml_getc(context);
                  switch (ch)
                    {
                    case 'A':
                      ch=doxml_getc(context);
                      switch (ch)
                        {
                        case '[':
                          return parse_cdata(context,parent);
                        }
                      break;
                    }
                  break;
                }
              break;
            }
          break;
        }
      break;
    }

  if (ch==EOF)
    {
      fprintf(stderr,"doxml: parse_banger(): reached EOF while trying to parse a <! directive\n");
      return 0;
    }

  fprintf(stderr,"doxml: parse_banger(): unrecognized <! directive\n");
  return 0;
}

static doxml_atom*
parse_pi(doxml_context* context,
         doxml_atom* parent)
{
  int ch;
  const char* name=0;
  const char* value=0;

  if (doxml_buff_init(context)<0)
    {
      fprintf(stderr,"doxml: parse_pi(): can't initialize buffer\n");
      return 0;
    }

  /* First, parse the name */
  while ((ch=doxml_getc(context))!=EOF)
    {
      if (isspace(ch))
        break;

      if (ch=='?')
        break;

      if (doxml_buff_append(context,ch)<0)
        {
          fprintf(stderr,"doxml: parse_pi(): can't append ch to buff\n");
          return 0;
        }
    }

  if (ch==EOF)
    {
      fprintf(stderr,"doxml: parse_pi(): EOF while parsing PITarget\n");
      return 0;
    }

  name=doxml_buff_extract(context);
  if (!name)
    {
      fprintf(stderr,"doxml: parse_pi(): can't extract PITarget from buffer\n");
      return 0;
    }

  if (ch=='?')
    {
      ch=doxml_getc(context);
      if (ch!='>')
        {
          fprintf(stderr,"doxml: parse_pi(): PITarget terminated by ?, but no >\n");
          return 0;
        }

      {
        doxml_atom* res=new_pi_atom(parent,0,0,name,0);
        if (!res)
          {
            free((char*)name);
            fprintf(stderr,"doxml: parse_pi(): new_pi_atom() failed\n");
            return 0;
          }
        
        return res;
      }
    }

  while (isspace(ch=doxml_getc(context)))
    ;

  if (ch==EOF)
    {
      free((char*)name);
      fprintf(stderr,"doxml: parse_pi(): EOF while parsing space\n");
      return 0;
    }

  if (doxml_buff_init(context)<0)
    {
      free((char*)name);
      fprintf(stderr,"doxml: parse_pi(): can't initialize buffer\n");
      return 0;
    }

  do {
    if (doxml_buff_append(context,ch)<0)
      {
        free((char*)name);
        fprintf(stderr,"doxml: parse_pi(): can't append ch to buff\n");
        return 0;
      }

    ch=doxml_getc(context);
    if (ch==EOF)
      break;

    if (ch=='?')
      {
        int tmp=doxml_getc(context);
        if (tmp!='>')
          doxml_ungetc(tmp,context);
        else
          ch='>';
      }
  } while (ch!='>');

  if (ch==EOF)
    {
      free((char*)name);
      fprintf(stderr,"doxml: parse_pi(): EOF in value\n");
      return 0;
    }

  value=doxml_buff_extract(context);
  if (!value)
    {
      free((char*)name);
      fprintf(stderr,"doxml: parse_pi(): can't extract value from buffer\n");
      return 0;
    }

  {
    doxml_atom* res=new_pi_atom(parent,0,0,name,value);
    if (!res)
      {
        free((char*)name);
        free((char*)value);
        fprintf(stderr,"doxml: parse_pi(): new_pi_atom() failed\n");
        return 0;
      }

    return res;
  }
}

static doxml_markupdecl*
get_attlist(const doxml_name* element_name,
            const doxml_doctype* doctype)
{
  doxml_markupdecl* cur;

  if (!doctype)
    return 0;

  if (!element_name)
    {
      fprintf(stderr,"doxml: get_attlist(): element_name is NULL\n");
      return 0;
    }

  for (cur=doctype->decls.first; cur; cur=cur->next)
    {
      if ((cur->type)!=doxml_markupdecl_attlist)
        continue;

      if (!doxml_tags_match_dtd(element_name,
                                &(cur->data.attlist.name)
                                )
          )
        continue;

      return cur;
    }

  return 0;
}

static doxml_markupdecl*
get_entity(const doxml_name_in_dtd* entity_name,
           const doxml_doctype* doctype)
{
  doxml_markupdecl* cur;

  if (!(entity_name && doctype))
    {
      fprintf(stderr,"doxml: get_attr_def(): entity_name or doctype is NULL\n");
      return 0;
    }

  for (cur=doctype->decls.first; cur; cur=cur->next)
    {
      if ((cur->type)!=doxml_markupdecl_entity)
        continue;

      if (!doxml_tags_match_dtd2(entity_name,
                                 &(cur->data.entity.name)
                                 )
          )
        continue;

      return cur;
    }

  return 0;
}

static doxml_markupdecl_attdef*
get_attr_def(const doxml_name* element_name,
             const char* name,
             const char* namespace_prefix,
             const doxml_doctype* doctype)
{
  doxml_markupdecl* decl;

  if (!(name && doctype))
    {
      fprintf(stderr,"doxml: get_attr_def(): name or doctype is NULL\n");
      return 0;
    }

  decl=get_attlist(element_name,doctype);
  if (!decl)
    return 0;

  {
    doxml_name attr_name;
    doxml_markupdecl_attdef* cur;

    attr_name.name=name;
    attr_name.space=0;
    attr_name.prefix=namespace_prefix;

    for (cur=decl->data.attlist.attdefs; cur; cur=cur->next)
      {
        if (!doxml_tags_match_dtd(&attr_name,
                                  &(cur->name)
                                  )
            )
          continue;

        return cur;
      }
  }

  return 0;
}

static int
attr_is_cdata(const doxml_name* element_name,
              const char* name,
              const char* namespace_prefix,
              const doxml_doctype* doctype)
{
  doxml_markupdecl_attdef* def=get_attr_def(element_name,
                                            name,namespace_prefix,
                                            doctype);
  if (!def)
    return 1;

  return (def->attype.type==doxml_markupdecl_attdef_attype_cdata);
}

static int
add_attr_maybe_entity(doxml_atom* atom,
                      const char* space,
                      const char* prefix,
                      const char* name,
                      const char* value,
                      const doxml_doctype* doctype)
{
  do {
    doxml_markupdecl_attdef* def
      =get_attr_def(&(atom->data.element.name),name,prefix,doctype);

    if (!def)
      break;
    if ((def->attype.type)!=doxml_markupdecl_attdef_attype_entity)
      break;

    {
      char* p=0;
      doxml_name_in_dtd entity_name;

      {
        const char* colon=strchr(value,':');
        if (colon)
          {
            p=(char*)malloc(colon-value+1);
            strncpy(p,value,colon-value);
            entity_name.prefix=p;
            entity_name.name=colon+1;
          }
        else
          {
            entity_name.prefix=0;
            entity_name.name=value;
          }
      }

      {
        doxml_markupdecl* entity=get_entity(&entity_name,doctype);
        if (entity)
          {
            int res=(doxml_attr_add_entity_with_prefix(atom,
                                                       space,prefix,name,
                                                       entity)!=0);
            if (p)
              free(p);
            return res ? 0 : -1;
          }
      }
    }
  } while(0);

  if (doxml_attr_add_with_prefix(atom,space,prefix,name,value))
    return 0;
  else
    return -1;
}

static doxml_atom*
parse_element_without_attr_defaults(doxml_context* context,
                                    doxml_atom* parent,
                                    int* null_means_error,
                                    const doxml_doctype* doctype)
{
  int ch,is_empty=0,is_terminator=0;
  const char* namespace_prefix=0;
  doxml_atom* atom=0;

#ifdef DEBUG
  fprintf(stderr,"doxml: parse_element_without_attr_defaults()\n");
#endif

  /* First, parse the tag */
  if (doxml_buff_init(context)<0)
    {
      fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't initialize buffer\n");
      return 0;
    }

 tag:
  while (!isspace(ch=doxml_getc(context)))
    {
      switch (ch)
        {
        case EOF:
          fprintf(stderr,"doxml: parse_element_without_attr_defaults(): EOF in element tag\n");
          return 0;

        case '/':
          is_empty=1;
          if (doxml_buff_is_empty(context))
            {
              if (doxml_buff_append(context,'/')<0)
                {
                  fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't append / to tag\n");
                  return 0;
                }
              is_terminator=1;
              continue;
            }
          ch=doxml_getc(context);
          if (ch!='>')
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): / not followed by >");
              return 0;
            }
          /* falling through on purpose */

        case '>':
          {
            const char* tag=doxml_buff_extract(context);
            if (!tag)
              {
                fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't extract tag from buffer\n");
                return 0;
              }
            if (namespace_prefix && (namespace_prefix[0]=='/'))
              {
                char* tag2=(char*)malloc(strlen(tag)+2);
                char* namespace_prefix2
                  =(char*)malloc(strlen(namespace_prefix));
                if (!(tag2 && namespace_prefix2))
                  {
                    fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't copy end tag\n");
                    if (tag2)
                      free(tag2);
                    if (namespace_prefix2)
                      free(namespace_prefix2);
                    return 0;
                  }

                tag2[0]='/';
                strcpy(tag2+1,tag);
                strcpy(namespace_prefix2,namespace_prefix+1);
                free((char*)tag);
                tag=tag2;
                free((char*)namespace_prefix);
                namespace_prefix=namespace_prefix2;
              }
            atom=new_element_atom(parent,
                                  doxml_lookup_namespace(parent,
                                                         namespace_prefix
                                                         ),
                                  namespace_prefix,
                                  tag);
            if (!atom)
              {
                fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't construct atom\n");
                return 0;
              }

            if (is_terminator)
              return atom;

            if (is_empty)
              return atom;
            else
              goto atoms;
          }
          break;

        case ':':
          if (namespace_prefix)
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): two namespace prefixes on same tag\n");
              return 0;
            }
          namespace_prefix=doxml_buff_extract(context);
          if (!namespace_prefix)
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't extract namespace_prefix\n");
              return 0;
            }
          if (doxml_buff_init(context)<0)
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't reinit buffer for tag\n");
              return 0;
            }
          break;

        default:
          if (doxml_buff_append(context,ch)<0)
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't append character to tag\n");
              return 0;
            }
          break;
        }
    }

 attrs:
  /* Skip whitespace between attrs */
  while (isspace(ch=doxml_getc(context)))
    ;

  if (!atom)
    {
      const char* tag=doxml_buff_extract(context);
      if (!tag)
        {
          fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't extract tag from buffer\n");
          return 0;
        }

      atom=new_element_atom(parent,
                            doxml_lookup_namespace(parent,
                                                   namespace_prefix
                                                   ),
                            namespace_prefix,
                            tag);
      if (!atom)
        {
          fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't construct atom\n");
          return 0;
        }
    }

  if (doxml_buff_init(context)<0)
    {
      fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't init buff for attr name\n");
      doxml_delete_atom(atom);
      return 0;
    }
  namespace_prefix=0;

  do
    {
      if (ch==EOF)
        {
          fprintf(stderr,"doxml: parse_element_without_attr_defaults(): EOF in an element\n");
          return 0;
        }

      if ((ch=='/') || (ch=='>'))
        {
          const char* name=doxml_buff_extract(context);
          if (!name)
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't extract attr name\n");
              doxml_delete_atom(atom);
              return 0;
            }

          if (name[0])
            {
              fprintf(stderr,
                      "doxml: parse_element_without_attr_defaults(): found attr \"%s\" with no value\n",
                      name);
              
              doxml_delete_atom(atom);
              return 0;
            }
        }

      if (ch=='=')
        goto value;

      if (ch=='/')
        {
          ch=doxml_getc(context);
          if (ch!='>')
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): / not followed by >");
              doxml_delete_atom(atom);
              return 0;
            }
          return atom;
        }

      if (ch=='>')
        goto atoms;

      if (ch==':')
        {
          if (namespace_prefix)
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): two namespace prefixes on same attr\n");
              return 0;
            }
          namespace_prefix=doxml_buff_extract(context);
          if (!namespace_prefix)
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't extract attr namespace_prefix\n");
              doxml_delete_atom(atom);
              return 0;
            }
          continue;
        }

      if (doxml_buff_append(context,ch)<0)
        {
          fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't append ch to attr name\n");
          doxml_delete_atom(atom);
          return 0;
        }
    }
  while (!isspace(ch=doxml_getc(context)));
  goto attrs;

 value:

  {
    const char* name=doxml_buff_extract(context);
    if (!name)
      {
        fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't extract attr name\n");
        doxml_delete_atom(atom);
        return 0;
      }

    {      
      const char* value=parse_quoted(context,doctype,1,
                                     !attr_is_cdata(&(atom->data.element.name),
                                                    name,namespace_prefix,
                                                    doctype),
                                     1
                                     );
      const char* attr_namespace=0;

      if (!value)
        {
          fprintf(stderr,"doxml: parse_element_without_attr_defaults(): could not extract attr value from buff\n");
          doxml_delete_atom(atom);
          return 0;
        }

      if ((!namespace_prefix) && (!strcmp(name,"xmlns")))
        {
          if (atom->data.element.name.space)
            free((char*)(atom->data.element.name.space));
          atom->data.element.name.space=strdup(value);
          if (!(atom->data.element.name.space))
            {
              fprintf(stderr,"doxml: parse_element_without_attr_defaults(): could not duplicate xmlns value\n");
              doxml_delete_atom(atom);
              return 0;
            }
          attr_namespace=0;
        }
      else
        attr_namespace=doxml_lookup_namespace(parent,
                                              namespace_prefix
                                              );

      if (add_attr_maybe_entity(atom,
                                attr_namespace,
                                namespace_prefix,
                                name,
                                value,
                                doctype
                                )
          <0
          )
        {
          fprintf(stderr,"doxml: parse_element_without_attr_defaults(): can't add attribute to element\n");
          doxml_delete_atom(atom);
          return 0;
        }
    }
  }

  goto attrs;

 atoms:
  {
    doxml_atom* child;
    while (child=doxml_parse_atom(context,atom,0,doctype,null_means_error))
      {
        if (child->data.element.name.name)
          {
            if (child->data.element.name.name[0]=='/')
              {
                if (doxml_tags_match(&(atom->data.element.name),
                                     &(child->data.element.name),
                                     1,1)
                    )
                  {
                    doxml_delete_atom(child);
                    return atom;
                  }
                else
                  {
                    fprintf(stderr,"doxml: parse_element_without_attr_defaults(): bad terminator: (%s,%s,%s) is terminated by (%s,%s,%s)\n",
                            safestr(atom->data.element.name.space),
                            safestr(atom->data.element.name.prefix),
                            safestr(atom->data.element.name.name),
                            safestr(child->data.element.name.space),
                            safestr(child->data.element.name.prefix),
                            safestr(child->data.element.name.name)
                            );
                    doxml_delete_atom(atom);
                    return 0;
                  }
              }
          }
        else
          {
            if (!(child->data.text.text))
              {
                fprintf(stderr,"doxml: parse_element_without_attr_defaults(): child atom has neither tag nor text\n");
                doxml_delete_atom(atom);
                return 0;
              }

            if (!(child->data.text.text[0]))
	      {
		fprintf(stderr,"doxml: %d: parse_element_without_attr_defaults(): !(child->data.text.text[0])",
			context->line_number);
		break;
	      }
          }
      }

    fprintf(stderr,"doxml: %d: parse_element_without_attr_defaults(): end of file while parsing children of (%s,%s)\n",
            context->line_number,
            safestr(atom->data.element.name.space),
            safestr(atom->data.element.name.name));
    doxml_delete_atom(atom);
    return 0;
  }
}

static doxml_atom*
parse_element(doxml_context* context,
              doxml_atom* parent,
              int* null_means_error,
              const doxml_doctype* doctype)
{
  doxml_markupdecl* attlist;
  doxml_atom* element=parse_element_without_attr_defaults(context,
                                                          parent,
                                                          null_means_error,
                                                          doctype);
  if (!element)
    return 0;

  /* Can't happen */
  if ((element->type)!=doxml_atom_element)
    return element;

  attlist=get_attlist(&(element->data.element.name),
                      doctype);
  if (!attlist)
    return element;

  /* Can't happen */
  if ((attlist->type)!=doxml_markupdecl_attlist)
    return element;

  {
    doxml_markupdecl_attdef* cur;
    for (cur=attlist->data.attlist.attdefs; cur; cur=cur->next)
      {
        if (!(cur->def.value))
          continue;

        if (doxml_attr_seek2(element,
                             &(cur->name)
                             )
            )
          continue;

        add_attr_maybe_entity(element,
                              0,cur->name.prefix,cur->name.name,
                              cur->def.value,
                              doctype);
      }
  }
}

static doxml_atom*
doxml_parse_atom(doxml_context* context,
                 doxml_atom* parent,
                 doxml_doctype** doctype_out,
                 const doxml_doctype* doctype_in,
                 int* null_means_error)
{
  int ch;
  int had_leading_whitespace=0;

  (*null_means_error)=1;

  if (context->external_entity)
    {
      const doxml_markupdecl* decl=context->external_entity;
      context->external_entity=0;
      return new_external_entity_atom(parent,decl);
    }

#ifdef DEBUG
  fprintf(stderr,"doxml: doxml_parse_atom()\n");
#endif

  if (!context)
    {
      fprintf(stderr,"doxml: doxml_parse(): context is NULL\n");
      return 0;
    }

  /* Skip leading whitespace */

  while (isspace(ch=doxml_getc(context)))
    had_leading_whitespace=1;

  if (ch==EOF)
    {
      (*null_means_error)=0;
      return 0;
    }

  if (ch=='<')                  /* It's an element or a PI */
    {
      ch=doxml_getc(context);
      switch (ch)
        {
        case '?':
          return parse_pi(context,parent);
        case '!':
          {
            int non_atom=0;
            doxml_atom* res=parse_banger(context,
                                         parent,
                                         doctype_out,
                                         doctype_in,
                                         &non_atom);
            if (non_atom)
              {
                if (res)
                  doxml_delete_atom(res);
                return doxml_parse_atom(context,parent,doctype_out,doctype_in,
                                        null_means_error);
              }
            else
              return res;
          }
          break;

        default:
          doxml_ungetc(ch,context);
          return parse_element(context,
                               parent,
                               null_means_error,
                               doctype_in);
        }
    }
  else
    {
      if (doxml_ungetc(ch,context)<0)
        {
          fprintf(stderr,"doxml: doxml_parse(): cannot ungetc first character of text span\n");
          return 0;
        }

      return parse_text(context,
                        parent,
                        had_leading_whitespace,
                        doctype_in);
    }
}

static int
starts_with(const char* s1,
            const char* s2)
{
  return !strncmp(s1,s2,strlen(s2));
}

static const char*
strndup(const char* s,
        int maxlen)
{
  int len=strlen(s);
  if (len>maxlen)
    len=maxlen;

  {
    char* res=(char*)malloc(len+1);
    if (res)
      {
        strncpy(res,s,len);
        res[len]=0;
      }
    return res;
  }
}

doxml_document*
doxml_parse(doxml_context* context)
{
  doxml_document* doc;
  doxml_atom* atom;
  int found_anything=0;
  int null_means_error;

#ifdef DEBUG
  fprintf(stderr,"doxml_parse()\n");
#endif

  doc=(doxml_document*)malloc(sizeof(doxml_document));
  if (!doc)
    {
      fprintf(stderr,"doxml_parse(): cannot malloc() doc\n");
      return 0;
    }

  doc->xml_decl.version=0;
  doc->xml_decl.encoding=0;
  doc->xml_decl.standalone=-1;
  doc->prolog=0;
  doc->doctype=0;
  doc->root=0;

  while (   (atom=doxml_parse_atom(context,0,
                                   &(doc->doctype),
                                   doc->doctype,
                                   &null_means_error))
         || (!null_means_error)
         )
    {
      if (atom->type==doxml_atom_element)
        {
          doc->root=atom;
          return doc;
        }

      if (atom->type==doxml_atom_pi)
        {
#ifdef DEBUG
          fprintf(stderr,"doxml_parse(): doxml_atom_pi\n");
#endif

          if ((!found_anything) && (!strcmp(atom->data.pi.name.name,"xml")))
            {
              const char* value=atom->data.pi.value;
              if (!value)
                continue;

#ifdef DEBUG
              fprintf(stderr,"doxml_parse(): <?xml?> value is %s\n",value);
#endif

              if (starts_with(value,"version"))
                {
                  const char* equals=strchr(value,'=');
                  const char* version=equals+1;
                  char quote;

                  while (isspace(version[0]))
                    version++;
                  
                  quote=version[0];
                  if (!quote)
                    continue;

                  version++;

                  {
                    const char* endquote=strchr(version,quote);
                    if (!endquote)
                      continue;

                    doc->xml_decl.version=strndup(version,(endquote-version));

                    value=endquote+1;
                    while (isspace(value[0]))
                      value++;
                  }
                }

              if (starts_with(value,"encoding"))
                {
                  const char* equals=strchr(value,'=');
                  const char* encoding=equals+1;
                  char quote;

                  while (isspace(encoding[0]))
                    encoding++;
                  
                  quote=encoding[0];
                  if (!quote)
                    continue;

                  encoding++;

                  {
                    const char* endquote=strchr(encoding,quote);
                    if (!endquote)
                      continue;

                    doc->xml_decl.encoding=strndup(encoding,(endquote-encoding));

                    value=endquote+1;
                    while (isspace(value[0]))
                      value++;
                  }
                }

              if (starts_with(value,"standalone"))
                {
                  const char* equals=strchr(value,'=');
                  const char* standalone=equals+1;
                  char quote;

                  while (isspace(standalone[0]))
                    standalone++;
                  
                  quote=standalone[0];
                  if (!quote)
                    continue;

                  standalone++;

                  {
                    char yes[5]="yes";
                    char no[4]="no";
                    yes[3]=quote;
                    yes[4]=0;
                    no[2]=quote;
                    no[3]=0;

                    if (starts_with(standalone,yes))
                      doc->xml_decl.standalone=1;
                    else
                      if (starts_with(standalone,no))
                        doc->xml_decl.standalone=0;
                  }

                  {
                    const char* endquote=strchr(standalone,quote);
                    if (!endquote)
                      continue;

                    value=endquote+1;
                    while (isspace(value[0]))
                      value++;
                  }
                }

              continue;
            }

          if (doc->prolog)
            {
              doxml_atom* cur;
              for (cur=doc->prolog; cur->next; cur=cur->next)
                ;

              cur->next=atom;
              atom->prev=cur;
            }
          else
            doc->prolog=atom;
        }

      found_anything=1;
    }

  if ((!found_anything) || null_means_error)
    {
      free(doc);
      return 0;
    }

  return doc;
}

int
doxml_doctype_handle_percent(doxml_context* context,
                             doxml_doctype* doctype)
{
  doxml_markupdecl* entity;

  doxml_name_in_dtd name;

  {
    int ch=doxml_getc(context);
    if (ch<0)
      return -1;
    doxml_ungetc(ch,context);
    if (isspace(ch))
      return 0;
  }

  if (parse_name_in_dtd(context,&name)<0)
    {
      fprintf(stderr,
              "doxml_doctype_handle_percent(): %d: can't parse entity-name\n",
              context->line_number);
      return -1;
    }

  {
    int ch=doxml_getc(context);
    if (ch!=';')
      {
        fprintf(stderr,
                "doxml_doctype_handle_percent(): %d: entity not terminated by ;\n",
                context->line_number);
        return -1;
      }
  }

  entity=get_entity(&name,doctype);
  if (!entity)
    {
      fprintf(stderr,
              "doxml_doctype_handle_percent(): %d: no such entity as ",
              context->line_number);
      if (name.prefix)
        fprintf(stderr,"%s:",name.prefix);
      if (name.name)
        fprintf(stderr,"%s",name.name);
      fprintf(stderr,"\n");
      return -1;
    }

  /* Can't happen */
  if (entity->type!=doxml_markupdecl_entity)
    {
      fprintf(stderr,
              "doxml_doctype_handle_percent(): %d: not an entity\n",
              context->line_number);
      return -1;
    }

  if (entity->data.entity.is_general)
    {
      fprintf(stderr,
              "doxml_doctype_handle_percent(): %d: not a param entity\n",
              context->line_number);
      return -1;
    }

  if (entity->data.entity.data.param.is_external)
    {
      fprintf(stderr,
              "doxml_doctype_handle_percent(): %d: external param entities not supported\n",
              context->line_number);
      return -1;
    }

  if (!(entity->data.entity.data.param.data.internal))
    {
      fprintf(stderr,
              "doxml_doctype_handle_percent(): %d: param entity with NULL internal\n",
              context->line_number);
      return -1;
    }

  {
    doxml_context* pushed
      =doxml_open_string(entity->data.entity.data.param.data.internal,0);
    if (!pushed)
      {
        fprintf(stderr,
                "doxml_doctype_handle_percent(): %d: can't open string\n",
                context->line_number);
        return -1;
      }

    if (doxml_context_push(context,pushed)<0)
      {
        fprintf(stderr,
                "doxml_doctype_handle_percent(): %d: can't push temporary context\n",
                context->line_number);
        return -1;
      }
  }

  return 1;
}
