/*
 * This is part of the doxml library, a parser for XML.  It should not
 *  require any functionality not defined in ANSI C's standard libraries.
 *
 * Copyright (C) 1999 John Stracke, http://www.thibault.org
 *
 * This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Log: utf8.c,v $
 * Revision 1.2  1999/07/18 22:36:58  francis
 * Adding support for character entities
 *
 * Revision 1.1  1999/07/18 22:00:20  francis
 * Adding basic Unicode-to-UTF8 conversion, needed for character entities
 *
 */

static const char* id="$Id: utf8.c,v 1.2 1999/07/18 22:36:58 francis Exp $";

#include <stdio.h>
#include <doxml/utf8.h>
#include <netinet/in.h>

static int
bytes_needed(const doxml_unicode_char* unicode)
{
  if ((unicode->bytes[0])>=4)
    return 6;
  if ((unicode->bytes[1])>=0x20)
    return 5;
  if ((unicode->bytes[1])>=1)
    return 4;
  if ((unicode->bytes[2])>=8)
    return 3;
  if ((unicode->bytes[3])>=0x80)
    return 2;
  return 1;
}

typedef union {
  unsigned char bytes[4];
  unsigned int w;
} u;

unsigned int doxml_word_host_to_network_order(unsigned int word)
{
  u out;
  out.bytes[0]=word>>24;
  out.bytes[1]=(word>>16)&255;
  out.bytes[2]=(word>>8)&255;
  out.bytes[3]=word&255;
  return out.w;
}

static unsigned int
word_network_to_host_order(unsigned int word)
{
  u in,out;
  u tst1,tst2;

  /* I would like to make tst1,tst2 static, but that would
   *  cause problems for multithreaded apps.
   */

  in.w=word;
  tst1.bytes[0]=0;
  tst1.bytes[1]=1;
  tst1.bytes[2]=2;
  tst1.bytes[3]=3;
  tst2.bytes[0]=tst1.w>>24;
  tst2.bytes[1]=(tst1.w>>16)&255;
  tst2.bytes[2]=(tst1.w>>8)&255;
  tst2.bytes[3]=(tst1.w)&255;

  out.w=(  ((in.bytes[tst2.bytes[3]])<<24)
         | ((in.bytes[tst2.bytes[2]])<<16)
         | ((in.bytes[tst2.bytes[1]])<<8)
         | (in.bytes[tst2.bytes[0]])
         );
  return out.w;
}

#if 0
int main(void)
{
  unsigned int words[]={57313,12,20892435,5,-5320105,0};
  int i,N;
  for (i=0,N=sizeof(words)/sizeof(words[0]); i<N; i++)
    {
      unsigned int h1=ntohl(words[i]);
      unsigned int h2=word_network_to_host_order(words[i]);
      printf("%x %x %x\n",words[i],h1,h2);
    }

  {
    unsigned int h=0x12345;
    unsigned int n1=htonl(h);
    unsigned int n2=doxml_word_host_to_network_order(h);
    printf("%x %x %x\n",h,n1,n2);
  }

  return 0;
}
#endif

static unsigned int
mask_low_n_bits(int n)
{
  unsigned int all=0xffffffff;
  return (all>>(32-n));
}

int
doxml_unicode_char_to_utf8(const doxml_unicode_char* unicode,
                           char* buff,
                           int bufflen)
{
  int nbytes;

  if (!unicode)
    {
      fprintf(stderr,"doxml_unicode_char_to_utf8(): unicode char is NULL\n");
      return -1;
    }

  if (!buff)
    {
      fprintf(stderr,"doxml_unicode_char_to_utf8(): buff is NULL\n");
      return -1;
    }

  nbytes=bytes_needed(unicode);
  if ((nbytes+1)>bufflen)
    {
      fprintf(stderr,"doxml_unicode_char_to_utf8(): bufflen is too small\n");
      return -1;
    }

  if (nbytes==1)
    buff[0]=unicode->bytes[3];
  else
    {
      unsigned int word=word_network_to_host_order(unicode->word);

      {
        int i;
        for (i=nbytes-1; i>0; i--)
          {
            buff[i]=(0x80 | (word&63));
            word>>=6;
          }
      }

      {
        int lead_bits=7-nbytes;
        buff[0]=(  ((0xfe<<(lead_bits))&0xff)
                 | word
                 );
      }
    }

  buff[nbytes]=0;
  return nbytes;
}
